try:  # run as a package if installed
    from pcntoolkit.normative_model.norm_blr import NormBLR
    from pcntoolkit.normative_model.norm_gpr import NormGPR
    from pcntoolkit.normative_model.norm_rfa import NormRFA
    from pcntoolkit.normative_model.norm_hbr import NormHBR
    from pcntoolkit.normative_model.norm_np import NormNP
except:
    from norm_blr import NormBLR
    from norm_gpr import NormGPR
    from norm_rfa import NormRFA
    from norm_hbr import NormHBR
    from norm_np import NormNP

def norm_init(X, y=None, theta=None, alg='gpr', **kwargs):
    if alg == 'gpr':
        nm = NormGPR(X=X, y=y, theta=theta, **kwargs)
    elif alg =='blr':
        nm = NormBLR(X=X, y=y, theta=theta, **kwargs)
    elif alg == 'rfa':
        nm = NormRFA(X=X, y=y, theta=theta, **kwargs)
    elif alg == 'hbr':
        nm = NormHBR(**kwargs)
    elif alg == 'np':
        nm = NormNP(X=X, y=y, **kwargs)
    else:
        raise(ValueError, "Algorithm " + alg + " not known.")
        
    return nm


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Nov 22 14:32:37 2019

@author: seykia
"""

import torch
from torch import nn
from torch.nn import functional as F

##################################### NP Model ################################

class NPR(nn.Module):
    def __init__(self, encoder, decoder, args):
        super(NPR, self).__init__()
        self.r_dim = encoder.r_dim
        self.z_dim = encoder.z_dim
        self.encoder = encoder
        self.decoder = decoder
        self.r_to_z_mean = nn.Linear(self.r_dim, self.z_dim)
        self.r_to_z_logvar = nn.Linear(self.r_dim, self.z_dim)
        self.device = args.device
        
    def xy_to_z_params(self, x, y):
        r = self.encoder.forward(x, y)
        mu = self.r_to_z_mean(r)
        logvar = self.r_to_z_logvar(r)
        return mu, logvar
      
    def reparameterise(self, z):
        mu, logvar = z
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        z_sample = eps.mul(std).add_(mu)
        return z_sample

    def forward(self, x_context, y_context, x_all=None, y_all=None, n = 10):
        y_sigma = None
        y_sigma_84 = None
        z_context = self.xy_to_z_params(x_context, y_context)
        if self.training:
            z_all = self.xy_to_z_params(x_all, y_all)
            z_sample = self.reparameterise(z_all)
            y_hat, y_hat_84 = self.decoder.forward(z_sample)
        else:  
            z_all = z_context
            temp = torch.zeros([n,y_context.shape[0], y_context.shape[2]], device = self.device)
            temp_84 = torch.zeros([n,y_context.shape[0], y_context.shape[2]], device = self.device)
            for i in range(n):
                z_sample = self.reparameterise(z_all)
                temp[i,:], temp_84[i,:] = self.decoder.forward(z_sample)
            y_hat = torch.mean(temp, dim=0).to(self.device)
            y_hat_84 = torch.mean(temp_84, dim=0).to(self.device)
            if n > 1:
                y_sigma = torch.std(temp, dim=0).to(self.device)
                y_sigma_84 = torch.std(temp_84, dim=0).to(self.device)
        return y_hat, y_hat_84, z_all, z_context, y_sigma, y_sigma_84
    
###############################################################################

def kl_div_gaussians(mu_q, logvar_q, mu_p, logvar_p):
    var_p = torch.exp(logvar_p)
    kl_div = (torch.exp(logvar_q) + (mu_q - mu_p) ** 2) / (var_p) \
             - 1.0 \
             + logvar_p - logvar_q
    kl_div = 0.5 * kl_div.sum()
    return kl_div

def np_loss(y_hat, y_hat_84, y, z_all, z_context):
    #PBL = pinball_loss(y, y_hat, 0.05)
    BCE = F.binary_cross_entropy(torch.squeeze(y_hat), torch.mean(y,dim=1), reduction="sum")
    idx1 = (y >= y_hat_84).squeeze()
    idx2 = (y < y_hat_84).squeeze()
    BCE84 = 0.84 * F.binary_cross_entropy(torch.squeeze(y_hat_84[idx1,:]), torch.mean(y[idx1,:],dim=1), reduction="sum") + \
            0.16 * F.binary_cross_entropy(torch.squeeze(y_hat_84[idx2,:]), torch.mean(y[idx2,:],dim=1), reduction="sum")
    KLD = kl_div_gaussians(z_all[0], z_all[1], z_context[0], z_context[1])
    return BCE + KLD + BCE84



from __future__ import print_function
from __future__ import division

import os
import sys
import numpy as np
import pandas as pd
from ast import literal_eval

try:  # run as a package if installed
    from pcntoolkit.model.bayesreg import BLR
    from pcntoolkit.normative_model.norm_base import NormBase
    from pcntoolkit.dataio import fileio
    from pcntoolkit.util.utils import create_poly_basis, WarpBoxCox, \
                                  WarpAffine, WarpCompose, WarpSinArcsinh
except ImportError:
    pass

    path = os.path.abspath(os.path.dirname(__file__))
    if path not in sys.path:
        sys.path.append(path)
    del path

    from model.bayesreg import BLR
    from norm_base import NormBase
    from dataio import fileio
    from util.utils import create_poly_basis, WarpBoxCox, \
                      WarpAffine, WarpCompose, WarpSinArcsinh

class NormBLR(NormBase):
    """ Normative modelling based on Bayesian Linear Regression
    """     
            
    def __init__(self,  **kwargs):
        X = kwargs.pop('X', None)
        y = kwargs.pop('y', None)
        theta = kwargs.pop('theta', None)
        if isinstance(theta, str):
            theta = np.array(literal_eval(theta))
        self.optim_alg = kwargs.get('optimizer','powell')

        if X is None:
            raise(ValueError, "Data matrix must be specified")

        if len(X.shape) == 1:
            self.D = 1
        else:
            self.D = X.shape[1]
        
        # Parse model order
        if kwargs is None:
            model_order = 1
        elif 'configparam' in kwargs: # deprecated syntax
            model_order = kwargs.pop('configparam')
        elif 'model_order' in kwargs: 
            model_order = kwargs.pop('model_order')
        else:
            model_order = 1
            
        # Force a default model order and check datatype
        if model_order is None:
            model_order = 1
        if type(model_order) is not int:
            model_order = int(model_order)
        
        # configure heteroskedastic noise
        if 'varcovfile' in kwargs:
            var_cov_file = kwargs.get('varcovfile')
            if var_cov_file.endswith('.pkl'):
                self.var_covariates = pd.read_pickle(var_cov_file)
            else:
                self.var_covariates = np.loadtxt(var_cov_file)
            if len(self.var_covariates.shape) == 1:
                self.var_covariates = self.var_covariates[:, np.newaxis]
            n_beta = self.var_covariates.shape[1]
            self.var_groups = None
        elif 'vargroupfile' in kwargs:
            # configure variance groups (e.g. site specific variance)
            var_groups_file = kwargs.pop('vargroupfile')
            if var_groups_file.endswith('.pkl'):
                self.var_groups = pd.read_pickle(var_groups_file)
            else:
                self.var_groups = np.loadtxt(var_groups_file)
            var_ids = set(self.var_groups)
            var_ids = sorted(list(var_ids))
            n_beta = len(var_ids)
        else:
            self.var_groups = None
            self.var_covariates = None
            n_beta = 1
        
        # are we using ARD?
        if 'use_ard' in kwargs: 
            self.use_ard = kwargs.pop('use_ard')
        else:
            self.use_ard = False
        if self.use_ard:
            n_alpha = self.D * model_order
        else:
            n_alpha = 1
        
        # Configure warped likelihood
        if 'warp' in kwargs:
            warp_str = kwargs.pop('warp')
            if warp_str is None:
                self.warp = None
                n_gamma = 0
            else:
                # set up warp
                exec('self.warp =' + warp_str + '()')
                n_gamma = self.warp.get_n_params()
        else:
            self.warp = None
            n_gamma = 0

        self._n_params = n_alpha + n_beta + n_gamma
        self._model_order = model_order
        
        print("configuring BLR ( order", model_order, ")")
        if (theta is None) or (len(theta) != self._n_params):
            print("Using default hyperparameters")
            self.theta0 = np.zeros(self._n_params)
        else:
            self.theta0 = theta
        self.theta = self.theta0
        
        # initialise the BLR object if the required parameters are present
        if (theta is not None) and (y is not None):
            Phi = create_poly_basis(X, self._model_order)
            self.blr = BLR(theta=theta, X=Phi, y=y, 
                           warp=self.warp, **kwargs)
        else:
            self.blr = BLR(**kwargs)    
            
    @property
    def n_params(self):
        return self._n_params
    
    @property
    def neg_log_lik(self):
        return self.blr.nlZ

    def estimate(self, X, y, **kwargs):
        theta = kwargs.pop('theta', None)
        if isinstance(theta, str):
            theta = np.array(literal_eval(theta))
            
        # remove warp string to prevent it being passed to the blr object
        kwargs.pop('warp',None) 
        
        Phi = create_poly_basis(X, self._model_order)
        if len(y.shape) > 1:
            y = y.ravel()
            
        if theta is None:
            theta = self.theta0           
            
            # (re-)initialize BLR object because parameters were not specified
            self.blr = BLR(theta=theta, X=Phi, y=y, 
                           var_groups=self.var_groups, 
                           warp=self.warp, **kwargs)

        self.theta = self.blr.estimate(theta, Phi, y, 
                                       var_covariates=self.var_covariates, **kwargs)
        
        return self

    def predict(self, Xs, X=None, y=None, **kwargs):      
        
        theta = self.theta # always use the estimated coefficients
        # remove from kwargs to avoid downstream problems
        kwargs.pop('theta', None)

        Phis = create_poly_basis(Xs, self._model_order)
        
        if X is None:
            Phi = None
        else:
            Phi = create_poly_basis(X, self._model_order)
        
        # process variance groups for the test data
        if 'testvargroup' in kwargs:
            var_groups_te = kwargs.pop('testvargroup')
        else:
            if 'testvargroupfile' in kwargs:
                var_groups_test_file = kwargs.pop('testvargroupfile')
                if var_groups_test_file.endswith('.pkl'):
                    var_groups_te = pd.read_pickle(var_groups_test_file)
                else:
                    var_groups_te = np.loadtxt(var_groups_test_file)
            else:
                var_groups_te = None
        
        # process test variance covariates
        if 'testvarcov' in kwargs:
            var_cov_te = kwargs.pop('testvarcov')
        else:
            if 'testvarcovfile' in kwargs:
                var_cov_test_file = kwargs.get('testvarcovfile')
                if var_cov_test_file.endswith('.pkl'):
                    var_cov_te = pd.read_pickle(var_cov_test_file)
                else:
                    var_cov_te = np.loadtxt(var_cov_test_file)
            else:
                var_cov_te = None
        
        # do we want to adjust the responses?
        if 'adaptresp' in kwargs:
            y_adapt = kwargs.pop('adaptresp')
        else:
            if 'adaptrespfile' in kwargs:
                y_adapt = fileio.load(kwargs.pop('adaptrespfile'))
                if len(y_adapt.shape) == 1:
                    y_adapt = y_adapt[:, np.newaxis]
            else:
                y_adapt = None
        
        if 'adaptcov' in kwargs:
            X_adapt = kwargs.pop('adaptcov')
            Phi_adapt = create_poly_basis(X_adapt, self._model_order)
        else:
            if 'adaptcovfile' in kwargs:
                X_adapt = fileio.load(kwargs.pop('adaptcovfile'))
                Phi_adapt = create_poly_basis(X_adapt, self._model_order)
            else:
                Phi_adapt = None
        
        if 'adaptvargroup' in kwargs:
            var_groups_ad = kwargs.pop('adaptvargroup')
        else:
            if 'adaptvargroupfile' in kwargs: 
                var_groups_adapt_file = kwargs.pop('adaptvargroupfile')
                if var_groups_adapt_file.endswith('.pkl'):
                    var_groups_ad = pd.read_pickle(var_groups_adapt_file)
                else:
                    var_groups_ad = np.loadtxt(var_groups_adapt_file)
            else:
                var_groups_ad = None
        
        if y_adapt is None:
            yhat, s2 = self.blr.predict(theta, Phi, y, Phis, 
                                        var_groups_test=var_groups_te,
                                        var_covariates_test=var_cov_te, 
                                        **kwargs)
        else:
            yhat, s2 = self.blr.predict_and_adjust(theta, Phi_adapt, y_adapt, Phis, 
                                                   var_groups_test=var_groups_te,
                                                   var_groups_adapt=var_groups_ad, 
                                                   **kwargs)
        
        return yhat, s2
    


#!/usr/bin/env python3

# -*- coding: utf-8 -*-
"""
Created on Tue Jun 18 09:47:01 2019

@author: seykia
"""
# ------------------------------------------------------------------------------
#  Usage:
#  python normative_NP.py -r /home/preclineu/andmar/data/seykia/ds000030_R1.0.5/responses.nii.gz 
#                           -c /home/preclineu/andmar/data/seykia/ds000030_R1.0.5/covariates.pickle 
#                           --tr /home/preclineu/andmar/data/seykia/ds000030_R1.0.5/test_responses.nii.gz 
#                           --tc /home/preclineu/andmar/data/seykia/ds000030_R1.0.5/test_covariates.pickle 
#                           -o /home/preclineu/andmar/data/seykia/ds000030_R1.0.5/Results
#
#
#  Written by S. M. Kia
# ------------------------------------------------------------------------------

from __future__ import print_function
from __future__ import division

import sys
import argparse
import torch
from torch import optim
import numpy as np
import pickle
from pcntoolkit.model.NP import NP, apply_dropout_test, np_loss
from sklearn.preprocessing import MinMaxScaler, StandardScaler
from sklearn.linear_model import LinearRegression, MultiTaskLasso
from pcntoolkit.model.architecture import Encoder, Decoder
from pcntoolkit.util.utils import compute_pearsonr, explained_var, compute_MSLL
from pcntoolkit.util.utils import extreme_value_prob, extreme_value_prob_fit, ravel_2D, unravel_2D
from pcntoolkit.dataio import fileio
import os

try:  # run as a package if installed
    from pcntoolkit import configs
except ImportError:
    pass

    path = os.path.abspath(os.path.dirname(__file__))
    if path not in sys.path:
        sys.path.append(path)
    del path
    import configs

def get_args(*args):
    """ Parse command line arguments"""

    ############################ Parsing inputs ###############################
    
    parser = argparse.ArgumentParser(description='Neural Processes (NP) for Deep Normative Modeling')
    parser.add_argument("-r", help="Training response nifti file address", 
                        required=True, dest="respfile", default=None)
    parser.add_argument("-c", help="Training covariates pickle file address", 
                        required=True, dest="covfile", default=None)
    parser.add_argument("--tc", help="Test covariates pickle file address", 
                        required=True, dest="testcovfile", default=None)
    parser.add_argument("--tr", help="Test response nifti file address", 
                        dest="testrespfile", default=None)
    parser.add_argument("--mask", help="Mask nifti file address", 
                        dest="mask", default=None)
    parser.add_argument("-o", help="Output directory address", dest="outdir", default=None)
    parser.add_argument('-m', type=int, default=10, dest='m',
                        help='number of fixed-effect estimations')
    parser.add_argument('--batchnum', type=int, default=10, dest='batchnum',
                        help='input batch size for training')
    parser.add_argument('--epochs', type=int, default=100, dest='epochs',
                        help='number of epochs to train')
    parser.add_argument('--device', type=str, default='cuda', dest='device',
                        help='Either cpu or cuda')
    parser.add_argument('--fxestimator', type=str, default='ST', dest='estimator',
                        help='Fixed-effect estimator type.')

    args = parser.parse_args()

    if (args.respfile == None or args.covfile == None or args.testcovfile == None):
        raise(ValueError, "Training response nifti file, Training covariates pickle file, and \
              Test covariates pickle file must be specified.")
    if (args.outdir == None):
        args.outdir = os.getcwd()
        
    cuda = args.device=='cuda' and torch.cuda.is_available()
    args.device = torch.device("cuda" if cuda else "cpu")
    args.kwargs = {'num_workers': 1, 'pin_memory': True} if cuda else {}
    args.type= 'MT'

    return args
    
def estimate(args):
    torch.set_default_dtype(torch.float32)
    args.type = 'MT'
    print('Loading the input Data ...')
    responses = fileio.load_nifti(args.respfile, vol=True).transpose([3,0,1,2])
    response_shape = responses.shape
    with open(args.covfile, 'rb') as handle:
        covariates = pickle.load(handle)['covariates'] 
    with open(args.testcovfile, 'rb') as handle:
        test_covariates = pickle.load(handle)['test_covariates']
    if args.mask is not None:
        mask = fileio.load_nifti(args.mask, vol=True)
        mask = fileio.create_mask(mask, mask=None)
    else: 
        mask = fileio.create_mask(responses[0,:,:,:], mask=None)
    if args.testrespfile is not None:
        test_responses = fileio.load_nifti(args.testrespfile, vol=True).transpose([3,0,1,2])
        test_responses_shape = test_responses.shape
    
    print('Normalizing the input Data ...')
    covariates_scaler = StandardScaler()
    covariates = covariates_scaler.fit_transform(covariates)
    test_covariates = covariates_scaler.transform(test_covariates)
    response_scaler = MinMaxScaler()
    responses = unravel_2D(response_scaler.fit_transform(ravel_2D(responses)), response_shape)
    if args.testrespfile is not None:
        test_responses = unravel_2D(response_scaler.transform(ravel_2D(test_responses)), test_responses_shape)
        test_responses = np.expand_dims(test_responses, axis=1)
    
    factor = args.m
    
    x_context = np.zeros([covariates.shape[0], factor, covariates.shape[1]], dtype=np.float32)
    y_context = np.zeros([responses.shape[0], factor, responses.shape[1], 
                         responses.shape[2], responses.shape[3]], dtype=np.float32)
    x_all = np.zeros([covariates.shape[0], factor, covariates.shape[1]], dtype=np.float32)
    x_context_test = np.zeros([test_covariates.shape[0], factor, test_covariates.shape[1]], dtype=np.float32)
    y_context_test = np.zeros([test_covariates.shape[0], factor, responses.shape[1], 
                         responses.shape[2], responses.shape[3]], dtype=np.float32)
    
    print('Estimating the fixed-effects ...')
    for i in range(factor):
        x_context[:,i,:] = covariates[:,:]
        x_context_test[:,i,:] = test_covariates[:,:]
        idx = np.random.randint(0,covariates.shape[0], covariates.shape[0])
        if args.estimator=='ST':
            for j in range(responses.shape[1]):
                for k in range(responses.shape[2]):
                    for l in range(responses.shape[3]):
                        reg = LinearRegression()
                        reg.fit(x_context[idx,i,:], responses[idx,j,k,l])
                        y_context[:,i,j,k,l] = reg.predict(x_context[:,i,:])
                        y_context_test[:,i,j,k,l] = reg.predict(x_context_test[:,i,:])
        elif args.estimator=='MT':
            reg = MultiTaskLasso(alpha=0.1)
            reg.fit(x_context[idx,i,:], np.reshape(responses[idx,:,:,:], [covariates.shape[0],np.prod(responses.shape[1:])]))
            y_context[:,i,:,:,:] = np.reshape(reg.predict(x_context[:,i,:]), 
                     [x_context.shape[0],responses.shape[1],responses.shape[2],responses.shape[3]])    
            y_context_test[:,i,:,:,:] = np.reshape(reg.predict(x_context_test[:,i,:]), 
                          [x_context_test.shape[0],responses.shape[1],responses.shape[2],responses.shape[3]])
        print('Fixed-effect %d of %d is computed!' %(i+1, factor))
    
    x_all = x_context
    responses = np.expand_dims(responses, axis=1).repeat(factor, axis=1)
    
    ################################## TRAINING #################################  
    
    encoder = Encoder(x_context, y_context, args).to(args.device)
    args.cnn_feature_num = encoder.cnn_feature_num
    decoder = Decoder(x_context, y_context, args).to(args.device)
    model = NP(encoder, decoder, args).to(args.device)
    
    print('Estimating the Random-effect ...')
    k = 1
    epochs = [int(args.epochs/4),int(args.epochs/2),int(args.epochs/5),int(args.epochs-args.epochs/4-args.epochs/2-args.epochs/5)]
    mini_batch_num = args.batchnum
    batch_size = int(x_context.shape[0]/mini_batch_num)
    model.train()
    for e in range(len(epochs)): 
        optimizer = optim.Adam(model.parameters(), lr=10**(-e-2))
        for j in range(epochs[e]):
            train_loss = 0
            rand_idx = np.random.permutation(x_context.shape[0])
            for i in range(mini_batch_num):
                optimizer.zero_grad()
                idx = rand_idx[i*batch_size:(i+1)*batch_size]
                y_hat, z_all, z_context, dummy = model(torch.tensor(x_context[idx,:,:], device = args.device), 
                                                   torch.tensor(y_context[idx,:,:,:,:], device = args.device), 
                                                   torch.tensor(x_all[idx,:,:], device = args.device), 
                                                   torch.tensor(responses[idx,:,:,:,:], device = args.device))
                loss = np_loss(y_hat, torch.tensor(responses[idx,:,:,:,:], device = args.device), z_all, z_context)
                loss.backward()
                train_loss += loss.item()
                optimizer.step()
            print('Epoch: %d, Loss:%f, Average Loss:%f' %(k, train_loss, train_loss/responses.shape[0]))
            k += 1
            
    ################################## Evaluation #################################
    
    print('Predicting on Test Data ...')
    model.eval()
    model.apply(apply_dropout_test)
    with torch.no_grad():
        y_hat, z_all, z_context, y_sigma = model(torch.tensor(x_context_test, device = args.device),
                                                 torch.tensor(y_context_test, device = args.device), n = 15)
    if args.testrespfile is not None:   
        test_loss = np_loss(y_hat[0:test_responses_shape[0],:], 
                            torch.tensor(test_responses, device = args.device), 
                            z_all, z_context).item()
        print('Average Test Loss:%f' %(test_loss/test_responses_shape[0]))
         
        RMSE = np.sqrt(np.mean((test_responses - y_hat[0:test_responses_shape[0],:].cpu().numpy())**2, axis = 0)).squeeze() * mask
        SMSE = RMSE ** 2 / np.var(test_responses, axis=0).squeeze()
        Rho, pRho = compute_pearsonr(test_responses.squeeze(), y_hat[0:test_responses_shape[0],:].cpu().numpy().squeeze())
        EXPV = explained_var(test_responses.squeeze(), y_hat[0:test_responses_shape[0],:].cpu().numpy().squeeze()) * mask
        MSLL = compute_MSLL(test_responses.squeeze(), y_hat[0:test_responses_shape[0],:].cpu().numpy().squeeze(), 
                            y_sigma[0:test_responses_shape[0],:].cpu().numpy().squeeze()**2, train_mean = test_responses.mean(0), 
                            train_var = test_responses.var(0)).squeeze() * mask
                            
        NPMs = (test_responses - y_hat[0:test_responses_shape[0],:].cpu().numpy()) / (y_sigma[0:test_responses_shape[0],:].cpu().numpy())
        NPMs = NPMs.squeeze()
        NPMs = NPMs * mask
        NPMs = np.nan_to_num(NPMs)
        
        temp=NPMs.reshape([NPMs.shape[0],NPMs.shape[1]*NPMs.shape[2]*NPMs.shape[3]])
        EVD_params = extreme_value_prob_fit(temp, 0.01)
        abnormal_probs = extreme_value_prob(EVD_params, temp, 0.01)
    
    ############################## SAVING RESULTS #################################

    print('Saving Results to: %s' %(args.outdir))
    exfile = args.respfile
    y_hat = y_hat.squeeze().cpu().numpy()
    y_hat = response_scaler.inverse_transform(ravel_2D(y_hat))
    y_hat = y_hat[:,mask.flatten()]
    fileio.save(y_hat.T, args.outdir + 
                '/yhat.nii.gz', example=exfile, mask=mask)
    ys2 = y_sigma.squeeze().cpu().numpy()
    ys2 = ravel_2D(ys2) * (response_scaler.data_max_ - response_scaler.data_min_)
    ys2 = ys2**2
    ys2 = ys2[:,mask.flatten()]
    fileio.save(ys2.T, args.outdir + 
                '/ys2.nii.gz', example=exfile, mask=mask)
    if args.testrespfile is not None:  
        NPMs = ravel_2D(NPMs)[:,mask.flatten()]
        fileio.save(NPMs.T, args.outdir +  
                    '/Z.nii.gz', example=exfile, mask=mask)
        fileio.save(Rho.flatten()[mask.flatten()], args.outdir +  
                    '/Rho.nii.gz', example=exfile, mask=mask)
        fileio.save(pRho.flatten()[mask.flatten()], args.outdir +  
                    '/pRho.nii.gz', example=exfile, mask=mask)
        fileio.save(RMSE.flatten()[mask.flatten()], args.outdir +  
                    '/rmse.nii.gz', example=exfile, mask=mask)
        fileio.save(SMSE.flatten()[mask.flatten()], args.outdir +  
                    '/smse.nii.gz', example=exfile, mask=mask)
        fileio.save(EXPV.flatten()[mask.flatten()], args.outdir +  
                    '/expv.nii.gz', example=exfile, mask=mask)
        fileio.save(MSLL.flatten()[mask.flatten()], args.outdir +  
                    '/msll.nii.gz', example=exfile, mask=mask)
    
    with open(args.outdir +'model.pkl', 'wb') as handle:
         pickle.dump({'model':model, 'covariates_scaler':covariates_scaler,
                      'response_scaler': response_scaler, 'EVD_params':EVD_params,
                      'abnormal_probs':abnormal_probs}, handle, protocol=configs.PICKLE_PROTOCOL)
    
###############################################################################
    print('DONE!')


def main(*args):
    """ Parse arguments and estimate model
    """
    
    np.seterr(invalid='ignore')
    args = get_args(args)
    estimate(args)
    
if __name__ == "__main__":
    main(sys.argv[1:])



#!/opt/conda/bin/python

# ------------------------------------------------------------------------------
#  Usage:
#  python trendsurf.py -m [maskfile] -b [basis] -c [covariates] <infile>
#
#  Written by A. Marquand
# ------------------------------------------------------------------------------

from __future__ import print_function
from __future__ import division

import os
import sys
import numpy as np
import nibabel as nib
import argparse

try:  # Run as a package if installed
    from pcntoolkit.dataio import fileio
    from pcntoolkit.model.bayesreg import BLR
except ImportError:
    pass
    path = os.path.abspath(os.path.dirname(__file__))
    if path not in sys.path:
        sys.path.append(path)
    del path

    from dataio import fileio
    from model.bayesreg import BLR



def load_data(datafile, maskfile=None):
    """ load 4d nifti data """
    if datafile.endswith("nii.gz") or datafile.endswith("nii"):
        # we load the data this way rather than fileio.load() because we need
        # access to the volumetric representation (to know the # coordinates)
        dat = fileio.load_nifti(datafile, vol=True)
        dim = dat.shape
        if len(dim) <= 3:
            dim = dim + (1,)
    else:
        raise ValueError("No routine to handle non-nifti data")

    mask = fileio.create_mask(dat, mask=maskfile)

    dat = fileio.vol2vec(dat, mask)
    maskid = np.where(mask.ravel())[0]

    # generate voxel coordinates
    i, j, k = np.meshgrid(np.linspace(0, dim[0]-1, dim[0]),
                          np.linspace(0, dim[1]-1, dim[1]),
                          np.linspace(0, dim[2]-1, dim[2]), indexing='ij')

    # voxel-to-world mapping
    img = nib.load(datafile)
    world = np.vstack((i.ravel(), j.ravel(), k.ravel(),
                       np.ones(np.prod(i.shape), float))).T
    world = np.dot(world, img.affine.T)[maskid, 0:3]

    return dat, world, mask


def create_basis(X, basis, mask):
    """ Create a (polynomial) basis set """

    # check whether we are using a polynomial basis set
    if type(basis) is int or (type(basis) is str and len(basis) == 1):
        dimpoly = int(basis)
        dimx = X.shape[1]
        print('Generating polynomial basis set of degree', dimpoly, '...')
        Phi = np.zeros((X.shape[0], X.shape[1]*dimpoly))
        colid = np.arange(0, dimx)
        for d in range(1, dimpoly+1):
            Phi[:, colid] = X ** d
            colid += dimx
    else:  # custom basis set
        if type(basis) is str:
            print('Loading custom basis set from', basis)

            # Phi_vol = fileio.load_data(basis)
            # we load the data this way instead so we can apply the same mask
            Phi_vol = fileio.load_nifti(basis, vol=True)
            Phi = fileio.vol2vec(Phi_vol, mask)
            print('Basis set consists of', Phi.shape[1], 'basis functions.')
            # maskid = np.where(mask.ravel())[0]
        else:
            raise ValueError("I don't know what to do with basis:", basis)

    return Phi


def write_nii(data, filename, examplenii, mask):
    """ Write output to nifti """

    # load example image
    ex_img = nib.load(examplenii)
    dim = ex_img.shape[0:3]
    nvol = int(data.shape[1])

    # write data
    array_data = np.zeros((np.prod(dim), nvol))
    array_data[mask.flatten(), :] = data
    array_data = np.reshape(array_data, dim+(nvol,))
    array_img = nib.Nifti1Image(array_data,
                                ex_img.get_affine(),
                                ex_img.get_header())
    nib.save(array_img, filename)


def get_args(*args):
    # parse arguments
    parser = argparse.ArgumentParser(description="Trend surface model")
    parser.add_argument("filename")
    parser.add_argument("-b", help="basis set", dest="basis", default=3)
    parser.add_argument("-m", help="mask file", dest="maskfile", default=None)
    parser.add_argument("-c", help="covariates file", dest="covfile",
                        default=None)
    parser.add_argument("-a", help="use ARD", action='store_true')
    parser.add_argument("-o", help="output all measures", action='store_true')
    args = parser.parse_args()
    wdir = os.path.realpath(os.path.curdir)
    filename = os.path.join(wdir, args.filename)
    if args.maskfile is None:
        maskfile = None
    else:
        maskfile = os.path.join(wdir, args.maskfile)
    basis = args.basis
    if args.covfile is not None:
        raise(NotImplementedError, "Covariates not implemented yet.")

    return filename, maskfile, basis, args.a, args.o


def estimate(filename, maskfile, basis, ard=False, outputall=False,
             saveoutput=True):
    """ Estimate a trend surface model

    This will estimate a trend surface model, independently for each subject.
    This is currently fit using a polynomial model of a specified degree.
    The models are estimated on the basis of data stored on disk in ascii or
    neuroimaging data formats (currently nifti only). Ascii data should be in
    tab or space delimited format with the number of voxels in rows and the
    number of subjects in columns. Neuroimaging data will be reshaped
    into the appropriate format

    Basic usage::

        estimate(filename, maskfile, basis)

    where the variables are defined below. Note that either the cfolds
    parameter or (testcov, testresp) should be specified, but not both.

    :param filename: 4-d nifti file containing the images to be estimated
    :param maskfile: nifti mask used to apply to the data
    :param basis: model order for the interpolating polynomial

    All outputs are written to disk in the same format as the input. These are:

    :outputs: * yhat - predictive mean
              * ys2 - predictive variance
              * trendcoeff - coefficients from the trend surface model
              * negloglik - Negative log marginal likelihood
              * hyp - hyperparameters
              * explainedvar - explained variance
              * rmse - standardised mean squared error
    """

    # load data
    print("Processing data in", filename)
    Y, X, mask = load_data(filename, maskfile)
    Y = np.round(10000*Y)/10000  # truncate precision to avoid numerical probs
    if len(Y.shape) == 1:
        Y = Y[:, np.newaxis]
    N = Y.shape[1]

    # standardize responses and covariates
    mY = np.mean(Y, axis=0)
    sY = np.std(Y, axis=0)
    Yz = (Y - mY) / sY
    mX = np.mean(X, axis=0)
    sX = np.std(X, axis=0)
    Xz = (X - mX) / sX

    # create basis set and set starting hyperparamters
    Phi = create_basis(Xz, basis, mask)
    if ard is True:
        hyp0 = np.zeros(Phi.shape[1]+1)
    else:
        hyp0 = np.zeros(2)

    # estimate the models for all subjects
    if ard:
        print('ARD is enabled')
    yhat = np.zeros_like(Yz)
    ys2 = np.zeros_like(Yz)
    nlZ = np.zeros(N)
    hyp = np.zeros((N, len(hyp0)))
    rmse = np.zeros(N)
    ev = np.zeros(N)
    m = np.zeros((N, Phi.shape[1]))
    bs2 = np.zeros((N, Phi.shape[1]))
    for i in range(0, N):
        print("Estimating model ", i+1, "of", N)
        breg = BLR()
        hyp[i, :] = breg.estimate(hyp0, Phi, Yz[:, i])
        m[i, :] = breg.m
        nlZ[i] = breg.nlZ

        # compute extra measures (e.g. marginal variances)?
        if outputall:
            bs2[i] = np.sqrt(np.diag(np.linalg.inv(breg.A)))

        # compute predictions and errors
        yhat[:, i], ys2[:, i] = breg.predict(hyp[i, :], Phi, Yz[:, i], Phi)
        yhat[:, i] = yhat[:, i]*sY[i] + mY[i]
        rmse[i] = np.sqrt(np.mean((Y[:, i] - yhat[:, i]) ** 2))
        ev[i] = 100*(1 - (np.var(Y[:, i] - yhat[:, i]) / np.var(Y[:, i])))

        print("Variance explained =", ev[i], "% RMSE =", rmse[i])

    print("Mean (std) variance explained =", ev.mean(), "(", ev.std(), ")")
    print("Mean (std) RMSE =", rmse.mean(), "(", rmse.std(), ")")

    # Write output
    if saveoutput:
        print("Writing output ...")
        np.savetxt("trendcoeff.txt", m, delimiter='\t', fmt='%5.8f')
        np.savetxt("negloglik.txt", nlZ, delimiter='\t', fmt='%5.8f')
        np.savetxt("hyp.txt", hyp, delimiter='\t', fmt='%5.8f')
        np.savetxt("explainedvar.txt", ev, delimiter='\t', fmt='%5.8f')
        np.savetxt("rmse.txt", rmse, delimiter='\t', fmt='%5.8f')
        fileio.save_nifti(yhat, 'yhat.nii.gz', filename, mask)
        fileio.save_nifti(ys2, 'ys2.nii.gz', filename, mask)

        if outputall:
            np.savetxt("trendcoeffvar.txt", bs2, delimiter='\t', fmt='%5.8f')
    else:
        out = [yhat, ys2, nlZ, hyp, rmse, ev, m]
        if outputall:
            out.append(bs2)
        return out

def main(*args):
    np.seterr(invalid='ignore')

    filename, maskfile, basis, ard, outputall = get_args(args)
    estimate(filename, maskfile, basis, ard, outputall)

# For running from the command line:
if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Nov 22 14:41:07 2019

@author: seykia
"""

from __future__ import print_function
from __future__ import division

import os
import sys
import numpy as np
import torch
from torch import nn, optim
from torch.nn import functional as F
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import MinMaxScaler
import pickle

try:  # run as a package if installed
    from pcntoolkit.normative_model.normbase import NormBase
    from pcntoolkit.model.NPR import NPR, np_loss
except ImportError:
    pass

    path = os.path.abspath(os.path.dirname(__file__))
    if path not in sys.path:
        sys.path.append(path)
    del path

    from model.NPR import NPR, np_loss
    from norm_base import NormBase

class struct(object):
    pass
   
class Encoder(nn.Module):
    def __init__(self, x, y, args):
        super(Encoder, self).__init__()
        self.r_dim = args.r_dim
        self.z_dim = args.z_dim
        self.hidden_neuron_num = args.hidden_neuron_num
        self.h_1 = nn.Linear(x.shape[1] + y.shape[1], self.hidden_neuron_num)
        self.h_2 = nn.Linear(self.hidden_neuron_num, self.hidden_neuron_num)
        self.h_3 = nn.Linear(self.hidden_neuron_num, self.r_dim)

    def forward(self, x, y):
        x_y = torch.cat([x, y], dim=2)
        x_y = F.relu(self.h_1(x_y))
        x_y = F.relu(self.h_2(x_y))
        x_y = F.relu(self.h_3(x_y))
        r = torch.mean(x_y, dim=1)
        return r
    
    
class Decoder(nn.Module):
    def __init__(self, x, y, args):
        super(Decoder, self).__init__()
        self.r_dim = args.r_dim
        self.z_dim = args.z_dim
        self.hidden_neuron_num = args.hidden_neuron_num
        
        self.g_1 = nn.Linear(self.z_dim, self.hidden_neuron_num)
        self.g_2 = nn.Linear(self.hidden_neuron_num, self.hidden_neuron_num)
        self.g_3 = nn.Linear(self.hidden_neuron_num, y.shape[1])
        
        self.g_1_84 = nn.Linear(self.z_dim, self.hidden_neuron_num)
        self.g_2_84 = nn.Linear(self.hidden_neuron_num, self.hidden_neuron_num)
        self.g_3_84 = nn.Linear(self.hidden_neuron_num, y.shape[1])
    
    def forward(self, z_sample):
        z_hat = F.relu(self.g_1(z_sample))
        z_hat = F.relu(self.g_2(z_hat))
        y_hat = torch.sigmoid(self.g_3(z_hat))
        
        z_hat_84 = F.relu(self.g_1(z_sample))
        z_hat_84 = F.relu(self.g_2_84(z_hat_84))
        y_hat_84 = torch.sigmoid(self.g_3_84(z_hat_84))
        
        return y_hat, y_hat_84
   


            
class NormNP(NormBase):
    """ Classical GPR-based normative modelling approach
    """
    
    def __init__(self, X, y, configparam=None):
        self.configparam = configparam
        if configparam is not None: 
            with open(configparam, 'rb') as handle:
                 config = pickle.load(handle)
            args = struct()
            if 'batch_size' in config:
                args.batch_size = config['batch_size']
            else:
                args.batch_size = 10
            if 'epochs' in config:
                args.epochs = config['epochs']
            else:
                args.epochs = 100
            if 'device' in config:
                args.device = config['device']
            else:
                args.device = torch.device('cpu')
            if 'm' in config:
                args.m = config['m']
            else:
                args.m = 200
            if 'hidden_neuron_num' in config:
                args.hidden_neuron_num = config['hidden_neuron_num']
            else:
                args.hidden_neuron_num = 10
            if 'r_dim' in config:
                args.r_dim = config['r_dim']
            else:
                args.r_dim = 5
            if 'z_dim' in config:
                args.z_dim = config['z_dim']
            else:
                args.z_dim = 3
            if 'nv' in config:
                args.nv = config['nv']
            else:
                args.nv = 0.01
        else:
            args = struct()
            args.batch_size = 10
            args.epochs = 100
            args.device = torch.device('cpu')
            args.m = 200
            args.hidden_neuron_num = 10
            args.r_dim = 5
            args.z_dim = 3
            args.nv = 0.01
        
        if y is not None:
            if y.ndim == 1:
                y = y.reshape(-1,1)
            self.args = args
            self.encoder = Encoder(X, y, args)
            self.decoder = Decoder(X, y, args)
            self.model = NPR(self.encoder, self.decoder, args)
       
        
    @property
    def n_params(self):
        return 1
    
    @property
    def neg_log_lik(self):
        return -1
    
    def estimate(self, X, y):
        if y.ndim == 1:
            y = y.reshape(-1,1)
        sample_num = X.shape[0]
        batch_size = self.args.batch_size
        factor_num = self.args.m
        mini_batch_num = int(np.floor(sample_num/batch_size))
        device = self.args.device
        
        self.scaler = MinMaxScaler()
        y = self.scaler.fit_transform(y)
 
        self.reg = []
        for i in range(factor_num):
            self.reg.append(LinearRegression())
            idx = np.random.randint(0, sample_num, sample_num)#int(sample_num/10))
            self.reg[i].fit(X[idx,:],y[idx,:])
        
        x_context = np.zeros([sample_num, factor_num, X.shape[1]])
        y_context = np.zeros([sample_num, factor_num, 1])
        
        s = X.std(axis=0)
        for j in range(factor_num):
            x_context[:,j,:] = X + np.sqrt(self.args.nv) * s * np.random.randn(X.shape[0], X.shape[1])
            y_context[:,j,:] = self.reg[j].predict(x_context[:,j,:])
        
        x_context = torch.tensor(x_context, device=device, dtype = torch.float)
        y_context = torch.tensor(y_context, device=device, dtype = torch.float)
        
        x_all = torch.tensor(np.expand_dims(X,axis=1), device=device, dtype = torch.float)
        y_all = torch.tensor(y.reshape(-1, 1, y.shape[1]), device=device, dtype = torch.float)        
        
        self.model.train()
        epochs = [int(self.args.epochs/4),int(self.args.epochs/2),int(self.args.epochs/5),
                  int(self.args.epochs-self.args.epochs/4-self.args.epochs/2-self.args.epochs/5)]
        k = 1
        for e in range(len(epochs)): 
            optimizer = optim.Adam(self.model.parameters(), lr=10**(-e-2))
            for j in range(epochs[e]):
                train_loss = 0
                for i in range(mini_batch_num):
                    optimizer.zero_grad()
                    idx = np.arange(i*batch_size,(i+1)*batch_size)
                    y_hat, y_hat_84, z_all, z_context, dummy, dummy = self.model(x_context[idx,:,:], y_context[idx,:,:], x_all[idx,:,:], y_all[idx,:,:])
                    loss = np_loss(y_hat, y_hat_84, y_all[idx,0,:], z_all, z_context)
                    loss.backward()
                    train_loss += loss.item()
                    optimizer.step()
                print('Epoch: %d, Loss:%f' %( k, train_loss))
                k += 1
        return self
        
    def predict(self, Xs, X=None, Y=None, theta=None): 
        sample_num = Xs.shape[0]
        factor_num = self.args.m
        x_context_test = np.zeros([sample_num, factor_num, Xs.shape[1]])
        y_context_test = np.zeros([sample_num, factor_num, 1])
        for j in range(factor_num):
            x_context_test[:,j,:] = Xs
            y_context_test[:,j,:] = self.reg[j].predict(x_context_test[:,j,:])
        x_context_test = torch.tensor(x_context_test, device=self.args.device, dtype = torch.float)
        y_context_test = torch.tensor(y_context_test, device=self.args.device, dtype = torch.float)
        self.model.eval()
        with torch.no_grad():
            y_hat, y_hat_84, z_all, z_context, y_sigma, y_sigma_84 = self.model(x_context_test, y_context_test, n = 100)
            
        y_hat = self.scaler.inverse_transform(y_hat.cpu().numpy())
        y_hat_84 = self.scaler.inverse_transform(y_hat_84.cpu().numpy())
        y_sigma = y_sigma.cpu().numpy() * (self.scaler.data_max_ - self.scaler.data_min_)
        y_sigma_84 = y_sigma_84.cpu().numpy() * (self.scaler.data_max_ - self.scaler.data_min_)
        sigma_al = y_hat - y_hat_84
        return y_hat.squeeze(), (y_sigma**2 + sigma_al**2).squeeze() #, z_context[0].cpu().numpy(), z_context[1].cpu().numpy()
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jul 25 13:23:15 2019

@author: seykia
@author: augub
"""

from __future__ import print_function
from __future__ import division
from ast import Param
from tkinter.font import names

import numpy as np
import pymc3 as pm
import theano
from itertools import product
from functools import reduce

import theano
from pymc3 import Metropolis, NUTS, Slice, HamiltonianMC
from scipy import stats
import bspline
from bspline import splinelab

from model.SHASH import SHASHo2, SHASHb, SHASHo
from util.utils import create_poly_basis
from util.utils import expand_all
from pcntoolkit.util.utils import cartesian_product

from theano import printing, function

def bspline_fit(X, order, nknots):
    feature_num = X.shape[1]
    bsp_basis = []

    for i in range(feature_num):
        minx = np.min(X[:,i])
        maxx = np.max(X[:,i])
        delta = maxx-minx
        # Expand range by 20% (10% on both sides)
        splinemin = minx-0.1*delta
        splinemax = maxx+0.1*delta
        knots = np.linspace(splinemin, splinemax, nknots)
        k = splinelab.augknt(knots, order)
        bsp_basis.append(bspline.Bspline(k, order))

    return bsp_basis

def bspline_transform(X, bsp_basis):
    if type(bsp_basis) != list:
        temp = []
        temp.append(bsp_basis)
        bsp_basis = temp

    feature_num = len(bsp_basis)
    X_transformed = []
    for f in range(feature_num):
        X_transformed.append(np.array([bsp_basis[f](i) for i in X[:, f]]))
    X_transformed = np.concatenate(X_transformed, axis=1)

    return X_transformed

def create_poly_basis(X, order):
    """ compute a polynomial basis expansion of the specified order"""

    if len(X.shape) == 1:
        X = X[:, np.newaxis]
    D = X.shape[1]
    Phi = np.zeros((X.shape[0], D * order))
    colid = np.arange(0, D)
    for d in range(1, order + 1):
        Phi[:, colid] = X ** d
        colid += D

    return Phi


def from_posterior(param, samples, distribution=None, half=False, freedom=1):
    if len(samples.shape) > 1:
        shape = samples.shape[1:]
    else:
        shape = None

    if (distribution is None):
        smin, smax = np.min(samples), np.max(samples)
        width = smax - smin
        x = np.linspace(smin, smax, 1000)
        y = stats.gaussian_kde(np.ravel(samples))(x)
        if half:
            x = np.concatenate([x, [x[-1] + 0.1 * width]])
            y = np.concatenate([y, [0]])
        else:
            x = np.concatenate([[x[0] - 0.1 * width], x, [x[-1] + 0.1 * width]])
            y = np.concatenate([[0], y, [0]])
        if shape is None:
            return pm.distributions.Interpolated(param, x, y)
        else:
            return pm.distributions.Interpolated(param, x, y, shape=shape)
    elif (distribution == 'normal'):
        temp = stats.norm.fit(samples)
        if shape is None:
            return pm.Normal(param, mu=temp[0], sigma=freedom * temp[1])
        else:
            return pm.Normal(param, mu=temp[0], sigma=freedom * temp[1], shape=shape)
    elif (distribution == 'hnormal'):
        temp = stats.halfnorm.fit(samples)
        if shape is None:
            return pm.HalfNormal(param, sigma=freedom * temp[1])
        else:
            return pm.HalfNormal(param, sigma=freedom * temp[1], shape=shape)
    elif (distribution == 'hcauchy'):
        temp = stats.halfcauchy.fit(samples)
        if shape is None:
            return pm.HalfCauchy(param, freedom * temp[1])
        else:
            return pm.HalfCauchy(param, freedom * temp[1], shape=shape)
    elif (distribution == 'uniform'):
        upper_bound = np.percentile(samples, 95)
        lower_bound = np.percentile(samples, 5)
        r = np.abs(upper_bound - lower_bound)
        if shape is None:
            return pm.Uniform(param, lower=lower_bound - freedom * r,
                              upper=upper_bound + freedom * r)
        else:
            return pm.Uniform(param, lower=lower_bound - freedom * r,
                              upper=upper_bound + freedom * r, shape=shape)
    elif (distribution == 'huniform'):
        upper_bound = np.percentile(samples, 95)
        lower_bound = np.percentile(samples, 5)
        r = np.abs(upper_bound - lower_bound)
        if shape is None:
            return pm.Uniform(param, lower=0, upper=upper_bound + freedom * r)
        else:
            return pm.Uniform(param, lower=0, upper=upper_bound + freedom * r, shape=shape)

    elif (distribution == 'gamma'):
        alpha_fit, loc_fit, invbeta_fit = stats.gamma.fit(samples)
        if shape is None:
            return pm.Gamma(param, alpha=freedom * alpha_fit, beta=freedom / invbeta_fit)
        else:
            return pm.Gamma(param, alpha=freedom * alpha_fit, beta=freedom / invbeta_fit, shape=shape)

    elif (distribution == 'igamma'):
        alpha_fit, loc_fit, beta_fit = stats.gamma.fit(samples)
        if shape is None:
            return pm.InverseGamma(param, alpha=freedom * alpha_fit, beta=freedom * beta_fit)
        else:
            return pm.InverseGamma(param, alpha=freedom * alpha_fit, beta=freedom * beta_fit, shape=shape)


def hbr(X, y, batch_effects, batch_effects_size, configs, trace=None):
    """
    :param X: [N×P] array of clinical covariates
    :param y: [N×1] array of neuroimaging measures
    :param batch_effects: [N×M] array of batch effects
    :param batch_effects_size: [b1, b2,...,bM] List of counts of unique values of batch effects
    :param configs:
    :param trace:
    :param return_shared_variables: If true, returns references to the shared variables. The values of the shared variables can be set manually, allowing running the same model on different data without re-compiling it. 
    :return:
    """
    X = theano.shared(X)
    X = theano.tensor.cast(X,'floatX')
    y = theano.shared(y)
    y = theano.tensor.cast(y,'floatX')


    with pm.Model() as model:

        # Make a param builder that will make the correct calls
        pb = ParamBuilder(model, X, y, batch_effects, trace, configs)

        if configs['likelihood'] == 'Normal':
            mu = pb.make_param("mu", mu_slope_mu_params = (0.,10.), 
                               sigma_slope_mu_params = (5.,), 
                               mu_intercept_mu_params=(0.,10.), 
                               sigma_intercept_mu_params = (5.,)).get_samples(pb)
            sigma = pb.make_param("sigma", mu_sigma_params = (10., 5.),
                                  sigma_sigma_params = (5.,)).get_samples(pb)
            sigma_plus = pm.math.log(1+pm.math.exp(sigma))
            y_like = pm.Normal('y_like',mu=mu, sigma=sigma_plus, observed=y)

        elif configs['likelihood'] in ['SHASHb','SHASHo','SHASHo2']:
            """
            Comment 1
            The current parameterizations are tuned towards standardized in- and output data. 
            It is possible to adjust the priors through the XXX_dist and XXX_params kwargs, like here we do with epsilon_params.
            Supported distributions are listed in the Prior class. 

            Comment 2
            Any mapping that is applied here after sampling should also be applied in util.hbr_utils.forward in order for the functions there to properly work. 
            For example, the softplus applied to sigma here is also applied in util.hbr_utils.forward
            """
            SHASH_map = {'SHASHb':SHASHb,'SHASHo':SHASHo,'SHASHo2':SHASHo2}

            mu =            pb.make_param("mu",         slope_mu_params = (0.,3.), mu_intercept_mu_params=(0.,1.), sigma_intercept_mu_params = (1.,)).get_samples(pb)
            sigma =         pb.make_param("sigma",      sigma_params = (1.,2.),    slope_sigma_params=(0.,1.),     intercept_sigma_params = (1., 1.)).get_samples(pb)
            sigma_plus =    pm.math.log(1+pm.math.exp(sigma))
            epsilon =       pb.make_param("epsilon",    epsilon_params = (0.,1.),  slope_epsilon_params=(0.,1.), intercept_epsilon_params=(0.,1)).get_samples(pb)
            delta =         pb.make_param("delta",      delta_params=(1.5,2.),     slope_delta_params=(0.,1),   intercept_delta_params=(2., 1)).get_samples(pb)
            delta_plus =    pm.math.log(1+pm.math.exp(delta)) + 0.3
            y_like = SHASH_map[configs['likelihood']]('y_like', mu=mu, sigma=sigma_plus, epsilon=epsilon, delta=delta_plus, observed = y)

    return model



class HBR:

    """Hierarchical Bayesian Regression for normative modeling

    Basic usage::

        model = HBR(configs)
        trace = model.estimate(X, y, batch_effects)
        ys,s2 = model.predict(X, batch_effects)

    where the variables are

    :param configs: a dictionary of model configurations.
    :param X: N-by-P input matrix of P features for N subjects
    :param y: N-by-1 vector of outputs.
    :param batch_effects: N-by-B matrix of B batch ids for N subjects.

    :returns: * ys - predictive mean
              * s2 - predictive variance

    Written by S.M. Kia
    """

    def get_step_methods(self, m):
        """
        This can be used to assign default step functions. However, the nuts initialization keyword doesnt work together with this... so better not use it. 

        STEP_METHODS = (
            NUTS,
            HamiltonianMC,
            Metropolis,
            BinaryMetropolis,
            BinaryGibbsMetropolis,
            Slice,
            CategoricalGibbsMetropolis,
        )
        :param m: a PyMC3 model
        :return:
        """
        samplermap = {'NUTS': NUTS, 'MH': Metropolis, 'Slice': Slice, 'HMC': HamiltonianMC}
        fallbacks = [Metropolis]         # We are using MH as a fallback method here
        if self.configs['sampler'] == 'NUTS':
            step_kwargs = {'nuts': {'target_accept': self.configs['target_accept']}}
        else:
            step_kwargs = None
        return pm.sampling.assign_step_methods(m, methods=[samplermap[self.configs['sampler']]] + fallbacks,
                                               step_kwargs=step_kwargs)

    def __init__(self, configs):
        self.bsp = None
        self.model_type = configs['type']
        self.configs = configs

    def get_modeler(self):
        return {'nn': nn_hbr}.get(self.model_type, hbr)
        
    def transform_X(self, X):
        if self.model_type == 'polynomial':
            Phi = create_poly_basis(X, self.configs['order'])
        elif self.model_type == 'bspline':
            if self.bsp is None:
                self.bsp = bspline_fit(X, self.configs['order'], self.configs['nknots'])
            bspline = bspline_transform(X, self.bsp)
            Phi = np.concatenate((X, bspline), axis = 1)
        else:
            Phi = X
        return Phi


    def find_map(self, X, y, batch_effects,method='L-BFGS-B'):
        """ Function to estimate the model """
        X, y, batch_effects = expand_all(X, y, batch_effects)

        self.batch_effects_num = batch_effects.shape[1]
        self.batch_effects_size = []
        for i in range(self.batch_effects_num):
            self.batch_effects_size.append(len(np.unique(batch_effects[:, i])))

        X = self.transform_X(X)
        modeler = self.get_modeler()
        with modeler(X, y, batch_effects, self.batch_effects_size, self.configs) as m:
            self.MAP = pm.find_MAP(method=method)
        return self.MAP

    def estimate(self, X, y, batch_effects):

        """ Function to estimate the model """
        X, y, batch_effects = expand_all(X, y, batch_effects)

        self.batch_effects_num = batch_effects.shape[1]
        self.batch_effects_size = []
        for i in range(self.batch_effects_num):
            self.batch_effects_size.append(len(np.unique(batch_effects[:, i])))

        X = self.transform_X(X)
        modeler = self.get_modeler()
        with modeler(X, y, batch_effects, self.batch_effects_size, self.configs) as m:
            self.trace = pm.sample(draws=self.configs['n_samples'],
                                   tune=self.configs['n_tuning'],
                                   chains=self.configs['n_chains'],
                                   init=self.configs['init'], n_init=500000,
                                   cores=self.configs['cores'])
        return self.trace

    def predict(self, X, batch_effects, pred='single'):
        """ Function to make predictions from the model """
        X, batch_effects = expand_all(X, batch_effects)

        samples = self.configs['n_samples']
        y = np.zeros([X.shape[0], 1])

        if pred == 'single':
            X = self.transform_X(X)
            modeler = self.get_modeler()
            with modeler(X, y, batch_effects, self.batch_effects_size, self.configs):
                ppc = pm.sample_posterior_predictive(self.trace, samples=samples, progressbar=True)
            pred_mean = ppc['y_like'].mean(axis=0)
            pred_var = ppc['y_like'].var(axis=0)

        return pred_mean, pred_var

    def estimate_on_new_site(self, X, y, batch_effects):
        """ Function to adapt the model """
        X, y, batch_effects = expand_all(X, y, batch_effects)

        self.batch_effects_num = batch_effects.shape[1]
        self.batch_effects_size = []
        for i in range(self.batch_effects_num):
            self.batch_effects_size.append(len(np.unique(batch_effects[:, i])))

        X = self.transform_X(X)
        modeler = self.get_modeler()
        with modeler(X, y, batch_effects, self.batch_effects_size,
                     self.configs, trace=self.trace) as m:
            self.trace = pm.sample(self.configs['n_samples'],
                                   tune=self.configs['n_tuning'],
                                   chains=self.configs['n_chains'],
                                   target_accept=self.configs['target_accept'],
                                   init=self.configs['init'], n_init=50000,
                                   cores=self.configs['cores'])
        return self.trace

    def predict_on_new_site(self, X, batch_effects):
        """ Function to make predictions from the model """
        X, batch_effects = expand_all(X, batch_effects)

        samples = self.configs['n_samples']
        y = np.zeros([X.shape[0], 1])

        X = self.transform_X(X)
        modeler = self.get_modeler()
        with modeler(X, y, batch_effects, self.batch_effects_size, self.configs, trace=self.trace):
            ppc = pm.sample_posterior_predictive(self.trace, samples=samples, progressbar=True)
        pred_mean = ppc['y_like'].mean(axis=0)
        pred_var = ppc['y_like'].var(axis=0)

        return pred_mean, pred_var

    def generate(self, X, batch_effects, samples):
        """ Function to generate samples from posterior predictive distribution """
        X, batch_effects = expand_all(X, batch_effects)

        y = np.zeros([X.shape[0], 1])

        X = self.transform_X(X)
        modeler = self.get_modeler()
        with modeler(X, y, batch_effects, self.batch_effects_size, self.configs):
            ppc = pm.sample_posterior_predictive(self.trace, samples=samples, progressbar=True)

        generated_samples = np.reshape(ppc['y_like'].squeeze().T, [X.shape[0] * samples, 1])
        X = np.repeat(X, samples)
        if len(X.shape) == 1:
            X = np.expand_dims(X, axis=1)
        batch_effects = np.repeat(batch_effects, samples, axis=0)
        if len(batch_effects.shape) == 1:
            batch_effects = np.expand_dims(batch_effects, axis=1)

        return X, batch_effects, generated_samples

    def sample_prior_predictive(self, X, batch_effects, samples, trace=None):
        """ Function to sample from prior predictive distribution """

        if len(X.shape) == 1:
            X = np.expand_dims(X, axis=1)
        if len(batch_effects.shape) == 1:
            batch_effects = np.expand_dims(batch_effects, axis=1)

        self.batch_effects_num = batch_effects.shape[1]
        self.batch_effects_size = []
        for i in range(self.batch_effects_num):
            self.batch_effects_size.append(len(np.unique(batch_effects[:, i])))

        y = np.zeros([X.shape[0], 1])

        if self.model_type == 'linear':
            with hbr(X, y, batch_effects, self.batch_effects_size, self.configs,
                     trace):
                ppc = pm.sample_prior_predictive(samples=samples)
        return ppc

    def get_model(self, X, y, batch_effects):
        X, y, batch_effects = expand_all(X, y, batch_effects)

        self.batch_effects_num = batch_effects.shape[1]
        self.batch_effects_size = []
        for i in range(self.batch_effects_num):
            self.batch_effects_size.append(len(np.unique(batch_effects[:, i])))
        modeler = self.get_modeler()
        X = self.transform_X(X)
        return modeler(X, y, batch_effects, self.batch_effects_size, self.configs, self.trace)

    def create_dummy_inputs(self, covariate_ranges = [[0.1,0.9,0.01]]):

        arrays = []
        for i in range(len(covariate_ranges)):
            arrays.append(np.arange(covariate_ranges[i][0],covariate_ranges[i][1], covariate_ranges[i][2]))
        X = cartesian_product(arrays)
        X_dummy = np.concatenate([X for i in range(np.prod(self.batch_effects_size))])

        arrays = []
        for i in range(self.batch_effects_num):
            arrays.append(np.arange(0, self.batch_effects_size[i]))
        batch_effects = cartesian_product(arrays)
        batch_effects_dummy = np.repeat(batch_effects, X.shape[0], axis=0)

        return X_dummy, batch_effects_dummy 

class Prior:
    """
    A wrapper class for a PyMC3 distribution. 
    - creates a fitted distribution from the trace, if one is present
    - overloads the __getitem__ function with something that switches between indexing or not, based on the shape
    """
    def __init__(self, name, dist, params, pb, shape=(1,)) -> None:
        self.dist = None
        self.name = name
        self.shape = shape
        self.has_random_effect = True if len(shape)>1 else False
        self.distmap = {'normal': pm.Normal,
                   'hnormal': pm.HalfNormal,
                   'gamma': pm.Gamma,
                   'uniform': pm.Uniform,
                   'igamma': pm.InverseGamma,
                   'hcauchy': pm.HalfCauchy}
        self.make_dist(dist, params, pb)
 
    def make_dist(self, dist, params, pb):
        """This creates a pymc3 distribution. If there is a trace, the distribution is fitted to the trace. If there isn't a trace, the prior is parameterized by the values in (params)"""
        with pb.model as m:
            if (pb.trace is not None) and (not self.has_random_effect):
                int_dist = from_posterior(param=self.name,
                                            samples=pb.trace[self.name],
                                            distribution=dist,
                                            freedom=pb.configs['freedom'])
                self.dist = int_dist.reshape(self.shape)
            else:
                shape_prod = np.product(np.array(self.shape))
                print(self.name)
                print(f"dist={dist}")
                print(f"params={params}")
                int_dist = self.distmap[dist](self.name, *params, shape=shape_prod)
                self.dist = int_dist.reshape(self.shape)

    def __getitem__(self, idx):
        """The idx here is the index of the batch-effect. If the prior does not model batch effects, this should return the same value for each index"""
        assert self.dist is not None, "Distribution not initialized"
        if self.has_random_effect:
            return self.dist[idx]
        else:
            return self.dist


class ParamBuilder:
    """
    A class that simplifies the construction of parameterizations. 
    It has a lot of attributes necessary for creating the model, including the data, but it is never saved with the model. 
    It also contains a lot of decision logic for creating the parameterizations.
    """

    def __init__(self, model, X, y, batch_effects, trace, configs):
        """

        :param model: model to attach all the distributions to
        :param X: Covariates
        :param y: IDPs
        :param batch_effects: I guess this speaks for itself
        :param trace:  idem
        :param configs: idem
        """
        self.model = model
        self.X = X
        self.y = y
        self.batch_effects = batch_effects
        self.trace = trace
        self.configs = configs

        self.feature_num = X.shape[1].eval().item()
        self.y_shape = y.shape.eval()
        self.n_ys = y.shape[0].eval().item()
        self.batch_effects_num = batch_effects.shape[1]

        self.batch_effects_size = []
        self.all_idx = []
        for i in range(self.batch_effects_num):
            # Count the unique values for each batch effect
            self.batch_effects_size.append(len(np.unique(self.batch_effects[:, i])))
            # Store the unique values for each batch effect
            self.all_idx.append(np.int16(np.unique(self.batch_effects[:, i])))
        # Make a cartesian product of all the unique values of each batch effect
        self.be_idx = list(product(*self.all_idx))

        # Make tuples of batch effects ID's and indices of datapoints with that specific combination of batch effects
        self.be_idx_tups = []
        for be in self.be_idx:
            a = []
            for i, b in enumerate(be):
                a.append(self.batch_effects[:, i] == b)
            idx = reduce(np.logical_and, a).nonzero()
            if idx[0].shape[0] != 0:
                self.be_idx_tups.append((be, idx))

    def make_param(self, name, dim = (1,), **kwargs):
        if self.configs.get(f'linear_{name}', False):
            # First make a slope and intercept, and use those to make a linear parameterization 
            slope_parameterization = self.make_param(f'slope_{name}', dim=[self.feature_num], **kwargs)
            intercept_parameterization = self.make_param(f'intercept_{name}', **kwargs)
            return LinearParameterization(name=name, dim=dim, 
                                    slope_parameterization=slope_parameterization, 
                                    intercept_parameterization=intercept_parameterization,
                                    pb=self, 
                                    **kwargs)
        
        elif self.configs.get(f'random_{name}', False):
            if self.configs.get(f'centered_{name}', True):
                return CentralRandomFixedParameterization(name=name, pb=self, dim=dim, **kwargs)
            else:
                return NonCentralRandomFixedParameterization(name=name, pb=self, dim=dim, **kwargs)
        else:
            return FixedParameterization(name=name, dim=dim, pb=self,**kwargs)


class Parameterization:
    """
    This is the top-level parameterization class from which all the other parameterizations inherit.
    """
    def __init__(self, name, dim):
        self.name = name
        self.dim = dim
        print(name, type(self))

    def get_samples(self, pb):

        with pb.model:
            samples = theano.tensor.zeros([pb.n_ys, *self.dim])
            for be, idx in pb.be_idx_tups:
                samples = theano.tensor.set_subtensor(samples[idx], self.dist[be])
        return samples


class FixedParameterization(Parameterization):
    """
    A parameterization that takes a single value for all input. It does not depend on anything except its hyperparameters
    """
    def __init__(self, name, dim, pb:ParamBuilder, **kwargs):
        super().__init__(name, dim)
        dist = kwargs.get(f'{name}_dist','normal')
        params = kwargs.get(f'{name}_params',(0.,1.))
        self.dist = Prior(name, dist, params, pb, shape = dim)


class CentralRandomFixedParameterization(Parameterization):
    """
    A parameterization that is fixed for each batch effect. This is sampled in a central fashion;
    the values are sampled from normal distribution with a group mean and group variance 
    """
    def __init__(self, name, dim, pb:ParamBuilder, **kwargs):
        super().__init__(name, dim)

        # Normal distribution is default for mean
        mu_dist = kwargs.get(f'mu_{name}_dist','normal')
        mu_params = kwargs.get(f'mu_{name}_params',(0.,1.))
        mu_prior = Prior(f'mu_{name}', mu_dist, mu_params, pb, shape = dim)

        # HalfCauchy is default for sigma
        sigma_dist = kwargs.get(f'sigma_{name}_dist','hcauchy')
        sigma_params = kwargs.get(f'sigma_{name}_params',(1.,))
        sigma_prior = Prior(f'sigma_{name}',sigma_dist, sigma_params, pb, shape = [*pb.batch_effects_size, *dim])

        self.dist = pm.Normal(name=name, mu=mu_prior.dist, sigma=sigma_prior.dist, shape = [*pb.batch_effects_size, *dim])
    

class NonCentralRandomFixedParameterization(Parameterization):
    """
    A parameterization that is fixed for each batch effect. This is sampled in a non-central fashion;
    the values are a sum of a group mean and noise values scaled with a group scaling factor 
    """
    def __init__(self, name,dim,  pb:ParamBuilder, **kwargs):
        super().__init__(name, dim)

        # Normal distribution is default for mean
        mu_dist = kwargs.get(f'mu_{name}_dist','normal')
        mu_params = kwargs.get(f'mu_{name}_params',(0.,1.))
        mu_prior = Prior(f'mu_{name}', mu_dist, mu_params, pb, shape = dim)

        # HalfCauchy is default for sigma
        sigma_dist = kwargs.get(f'sigma_{name}_dist','hcauchy')
        sigma_params = kwargs.get(f'sigma_{name}_params',(1.,))
        sigma_prior = Prior(f'sigma_{name}',sigma_dist, sigma_params, pb, shape = dim)

        # Normal is default for offset
        offset_dist = kwargs.get(f'offset_{name}_dist','normal')
        offset_params = kwargs.get(f'offset_{name}_params',(0.,1.))
        offset_prior = Prior(f'offset_{name}',offset_dist, offset_params, pb, shape = [*pb.batch_effects_size, *dim])

        self.dist = pm.Deterministic(name=name, var=mu_prior.dist+sigma_prior.dist*offset_prior.dist)


class LinearParameterization(Parameterization):
    """
    A parameterization that can model a linear dependence on X. 
    """
    def __init__(self, name, dim, slope_parameterization, intercept_parameterization, pb, **kwargs):
        super().__init__( name, dim)
        self.slope_parameterization = slope_parameterization
        self.intercept_parameterization = intercept_parameterization

    def get_samples(self, pb:ParamBuilder):
        with pb.model:
            samples = theano.tensor.zeros([pb.n_ys, *self.dim])
            for be, idx in pb.be_idx_tups:
                dot = theano.tensor.dot(pb.X[idx,:], self.slope_parameterization.dist[be]).T
                intercept = self.intercept_parameterization.dist[be]
                samples = theano.tensor.set_subtensor(samples[idx,:],dot+intercept)
        return samples


def get_design_matrix(X, nm, basis="linear"):
    if basis == "bspline":
        Phi = bspline_transform(X, nm.hbr.bsp)
    elif basis == "polynomial":
        Phi = create_poly_basis(X, 3)
    else:
        Phi = X
    return Phi



def nn_hbr(X, y, batch_effects, batch_effects_size, configs, trace=None):
    n_hidden = configs['nn_hidden_neuron_num']
    n_layers = configs['nn_hidden_layers_num']
    feature_num = X.shape[1]
    batch_effects_num = batch_effects.shape[1]
    all_idx = []
    for i in range(batch_effects_num):
        all_idx.append(np.int16(np.unique(batch_effects[:, i])))
    be_idx = list(product(*all_idx))

    # Initialize random weights between each layer for the mu:
    init_1 = pm.floatX(np.random.randn(feature_num, n_hidden) * np.sqrt(1 / feature_num))
    init_out = pm.floatX(np.random.randn(n_hidden) * np.sqrt(1 / n_hidden))

    std_init_1 = pm.floatX(np.random.rand(feature_num, n_hidden))
    std_init_out = pm.floatX(np.random.rand(n_hidden))

    # And initialize random weights between each layer for sigma_noise:
    init_1_noise = pm.floatX(np.random.randn(feature_num, n_hidden) * np.sqrt(1 / feature_num))
    init_out_noise = pm.floatX(np.random.randn(n_hidden) * np.sqrt(1 / n_hidden))

    std_init_1_noise = pm.floatX(np.random.rand(feature_num, n_hidden))
    std_init_out_noise = pm.floatX(np.random.rand(n_hidden))

    # If there are two hidden layers, then initialize weights for the second layer:
    if n_layers == 2:
        init_2 = pm.floatX(np.random.randn(n_hidden, n_hidden) * np.sqrt(1 / n_hidden))
        std_init_2 = pm.floatX(np.random.rand(n_hidden, n_hidden))
        init_2_noise = pm.floatX(np.random.randn(n_hidden, n_hidden) * np.sqrt(1 / n_hidden))
        std_init_2_noise = pm.floatX(np.random.rand(n_hidden, n_hidden))

    with pm.Model() as model:

        X = pm.Data('X', X)
        y = pm.Data('y', y)

        if trace is not None:  # Used when estimating/predicting on a new site
            weights_in_1_grp = from_posterior('w_in_1_grp', trace['w_in_1_grp'],
                                              distribution='normal', freedom=configs['freedom'])

            weights_in_1_grp_sd = from_posterior('w_in_1_grp_sd', trace['w_in_1_grp_sd'],
                                                 distribution='hcauchy', freedom=configs['freedom'])

            if n_layers == 2:
                weights_1_2_grp = from_posterior('w_1_2_grp', trace['w_1_2_grp'],
                                                 distribution='normal', freedom=configs['freedom'])

                weights_1_2_grp_sd = from_posterior('w_1_2_grp_sd', trace['w_1_2_grp_sd'],
                                                    distribution='hcauchy', freedom=configs['freedom'])

            weights_2_out_grp = from_posterior('w_2_out_grp', trace['w_2_out_grp'],
                                               distribution='normal', freedom=configs['freedom'])

            weights_2_out_grp_sd = from_posterior('w_2_out_grp_sd', trace['w_2_out_grp_sd'],
                                                  distribution='hcauchy', freedom=configs['freedom'])

            mu_prior_intercept = from_posterior('mu_prior_intercept', trace['mu_prior_intercept'],
                                                distribution='normal', freedom=configs['freedom'])
            sigma_prior_intercept = from_posterior('sigma_prior_intercept', trace['sigma_prior_intercept'],
                                                   distribution='hcauchy', freedom=configs['freedom'])

        else:
            # Group the mean distribution for input to the hidden layer:
            weights_in_1_grp = pm.Normal('w_in_1_grp', 0, sd=1,
                                         shape=(feature_num, n_hidden), testval=init_1)

            # Group standard deviation:
            weights_in_1_grp_sd = pm.HalfCauchy('w_in_1_grp_sd', 1.,
                                                shape=(feature_num, n_hidden), testval=std_init_1)

            if n_layers == 2:
                # Group the mean distribution for hidden layer 1 to hidden layer 2:
                weights_1_2_grp = pm.Normal('w_1_2_grp', 0, sd=1,
                                            shape=(n_hidden, n_hidden), testval=init_2)

                # Group standard deviation:
                weights_1_2_grp_sd = pm.HalfCauchy('w_1_2_grp_sd', 1.,
                                                   shape=(n_hidden, n_hidden), testval=std_init_2)

            # Group the mean distribution for hidden to output:
            weights_2_out_grp = pm.Normal('w_2_out_grp', 0, sd=1,
                                          shape=(n_hidden,), testval=init_out)

            # Group standard deviation:
            weights_2_out_grp_sd = pm.HalfCauchy('w_2_out_grp_sd', 1.,
                                                 shape=(n_hidden,), testval=std_init_out)

            # mu_prior_intercept = pm.Uniform('mu_prior_intercept', lower=-100, upper=100)
            mu_prior_intercept = pm.Normal('mu_prior_intercept', mu=0., sigma=1e3)
            sigma_prior_intercept = pm.HalfCauchy('sigma_prior_intercept', 5)

        # Now create separate weights for each group, by doing
        # weights * group_sd + group_mean, we make sure the new weights are
        # coming from the (group_mean, group_sd) distribution.
        weights_in_1_raw = pm.Normal('w_in_1', 0, sd=1,
                                     shape=(batch_effects_size + [feature_num, n_hidden]))
        weights_in_1 = weights_in_1_raw * weights_in_1_grp_sd + weights_in_1_grp

        if n_layers == 2:
            weights_1_2_raw = pm.Normal('w_1_2', 0, sd=1,
                                        shape=(batch_effects_size + [n_hidden, n_hidden]))
            weights_1_2 = weights_1_2_raw * weights_1_2_grp_sd + weights_1_2_grp

        weights_2_out_raw = pm.Normal('w_2_out', 0, sd=1,
                                      shape=(batch_effects_size + [n_hidden]))
        weights_2_out = weights_2_out_raw * weights_2_out_grp_sd + weights_2_out_grp

        intercepts_offset = pm.Normal('intercepts_offset', mu=0, sd=1,
                                      shape=(batch_effects_size))

        intercepts = pm.Deterministic('intercepts', intercepts_offset +
                                      mu_prior_intercept * sigma_prior_intercept)

        # Build the neural network and estimate y_hat:
        y_hat = theano.tensor.zeros(y.shape)
        for be in be_idx:
            # Find the indices corresponding to 'group be':
            a = []
            for i, b in enumerate(be):
                a.append(batch_effects[:, i] == b)
            idx = reduce(np.logical_and, a).nonzero()
            if idx[0].shape[0] != 0:
                act_1 = pm.math.tanh(theano.tensor.dot(X[idx, :], weights_in_1[be]))
                if n_layers == 2:
                    act_2 = pm.math.tanh(theano.tensor.dot(act_1, weights_1_2[be]))
                    y_hat = theano.tensor.set_subtensor(y_hat[idx, 0],
                                                        intercepts[be] + theano.tensor.dot(act_2, weights_2_out[be]))
                else:
                    y_hat = theano.tensor.set_subtensor(y_hat[idx, 0],
                                                        intercepts[be] + theano.tensor.dot(act_1, weights_2_out[be]))

        # If we want to estimate varying noise terms across groups:
        if configs['random_noise']:
            if configs['hetero_noise']:
                if trace is not None:  # # Used when estimating/predicting on a new site
                    weights_in_1_grp_noise = from_posterior('w_in_1_grp_noise',
                                                            trace['w_in_1_grp_noise'],
                                                            distribution='normal', freedom=configs['freedom'])

                    weights_in_1_grp_sd_noise = from_posterior('w_in_1_grp_sd_noise',
                                                               trace['w_in_1_grp_sd_noise'],
                                                               distribution='hcauchy', freedom=configs['freedom'])

                    if n_layers == 2:
                        weights_1_2_grp_noise = from_posterior('w_1_2_grp_noise',
                                                               trace['w_1_2_grp_noise'],
                                                               distribution='normal', freedom=configs['freedom'])

                        weights_1_2_grp_sd_noise = from_posterior('w_1_2_grp_sd_noise',
                                                                  trace['w_1_2_grp_sd_noise'],
                                                                  distribution='hcauchy', freedom=configs['freedom'])

                    weights_2_out_grp_noise = from_posterior('w_2_out_grp_noise',
                                                             trace['w_2_out_grp_noise'],
                                                             distribution='normal', freedom=configs['freedom'])

                    weights_2_out_grp_sd_noise = from_posterior('w_2_out_grp_sd_noise',
                                                                trace['w_2_out_grp_sd_noise'],
                                                                distribution='hcauchy', freedom=configs['freedom'])

                else:
                    # The input layer to the first hidden layer:
                    weights_in_1_grp_noise = pm.Normal('w_in_1_grp_noise', 0, sd=1,
                                                       shape=(feature_num, n_hidden),
                                                       testval=init_1_noise)
                    weights_in_1_grp_sd_noise = pm.HalfCauchy('w_in_1_grp_sd_noise', 1,
                                                              shape=(feature_num, n_hidden),
                                                              testval=std_init_1_noise)

                    # The first hidden layer to second hidden layer:
                    if n_layers == 2:
                        weights_1_2_grp_noise = pm.Normal('w_1_2_grp_noise', 0, sd=1,
                                                          shape=(n_hidden, n_hidden),
                                                          testval=init_2_noise)
                        weights_1_2_grp_sd_noise = pm.HalfCauchy('w_1_2_grp_sd_noise', 1,
                                                                 shape=(n_hidden, n_hidden),
                                                                 testval=std_init_2_noise)

                    # The second hidden layer to output layer:
                    weights_2_out_grp_noise = pm.Normal('w_2_out_grp_noise', 0, sd=1,
                                                        shape=(n_hidden,),
                                                        testval=init_out_noise)
                    weights_2_out_grp_sd_noise = pm.HalfCauchy('w_2_out_grp_sd_noise', 1,
                                                               shape=(n_hidden,),
                                                               testval=std_init_out_noise)

                    # mu_prior_intercept_noise = pm.HalfNormal('mu_prior_intercept_noise', sigma=1e3)
                    # sigma_prior_intercept_noise = pm.HalfCauchy('sigma_prior_intercept_noise', 5)

                # Now create separate weights for each group:
                weights_in_1_raw_noise = pm.Normal('w_in_1_noise', 0, sd=1,
                                                   shape=(batch_effects_size + [feature_num, n_hidden]))
                weights_in_1_noise = weights_in_1_raw_noise * weights_in_1_grp_sd_noise + weights_in_1_grp_noise

                if n_layers == 2:
                    weights_1_2_raw_noise = pm.Normal('w_1_2_noise', 0, sd=1,
                                                      shape=(batch_effects_size + [n_hidden, n_hidden]))
                    weights_1_2_noise = weights_1_2_raw_noise * weights_1_2_grp_sd_noise + weights_1_2_grp_noise

                weights_2_out_raw_noise = pm.Normal('w_2_out_noise', 0, sd=1,
                                                    shape=(batch_effects_size + [n_hidden]))
                weights_2_out_noise = weights_2_out_raw_noise * weights_2_out_grp_sd_noise + weights_2_out_grp_noise

                # intercepts_offset_noise = pm.Normal('intercepts_offset_noise', mu=0, sd=1,
                #                          shape=(batch_effects_size))

                # intercepts_noise = pm.Deterministic('intercepts_noise', mu_prior_intercept_noise +
                #                      intercepts_offset_noise * sigma_prior_intercept_noise)

                # Build the neural network and estimate the sigma_y:
                sigma_y = theano.tensor.zeros(y.shape)
                for be in be_idx:
                    a = []
                    for i, b in enumerate(be):
                        a.append(batch_effects[:, i] == b)
                    idx = reduce(np.logical_and, a).nonzero()
                    if idx[0].shape[0] != 0:
                        act_1_noise = pm.math.sigmoid(theano.tensor.dot(X[idx, :], weights_in_1_noise[be]))
                        if n_layers == 2:
                            act_2_noise = pm.math.sigmoid(theano.tensor.dot(act_1_noise, weights_1_2_noise[be]))
                            temp = pm.math.log1pexp(theano.tensor.dot(act_2_noise, weights_2_out_noise[be])) + 1e-5
                        else:
                            temp = pm.math.log1pexp(theano.tensor.dot(act_1_noise, weights_2_out_noise[be])) + 1e-5
                        sigma_y = theano.tensor.set_subtensor(sigma_y[idx, 0], temp)

            else:  # homoscedastic noise:
                if trace is not None:  # Used for transferring the priors
                    upper_bound = np.percentile(trace['sigma_noise'], 95)
                    sigma_noise = pm.Uniform('sigma_noise', lower=0, upper=2 * upper_bound, shape=(batch_effects_size))
                else:
                    sigma_noise = pm.Uniform('sigma_noise', lower=0, upper=100, shape=(batch_effects_size))

                sigma_y = theano.tensor.zeros(y.shape)
                for be in be_idx:
                    a = []
                    for i, b in enumerate(be):
                        a.append(batch_effects[:, i] == b)
                    idx = reduce(np.logical_and, a).nonzero()
                    if idx[0].shape[0] != 0:
                        sigma_y = theano.tensor.set_subtensor(sigma_y[idx, 0], sigma_noise[be])

        else:  # do not allow for random noise terms across groups:
            if trace is not None:  # Used for transferring the priors
                upper_bound = np.percentile(trace['sigma_noise'], 95)
                sigma_noise = pm.Uniform('sigma_noise', lower=0, upper=2 * upper_bound)
            else:
                sigma_noise = pm.Uniform('sigma_noise', lower=0, upper=100)
            sigma_y = theano.tensor.zeros(y.shape)
            for be in be_idx:
                a = []
                for i, b in enumerate(be):
                    a.append(batch_effects[:, i] == b)
                idx = reduce(np.logical_and, a).nonzero()
                if idx[0].shape[0] != 0:
                    sigma_y = theano.tensor.set_subtensor(sigma_y[idx, 0], sigma_noise)

        if configs['skewed_likelihood']:
            skewness = pm.Uniform('skewness', lower=-10, upper=10, shape=(batch_effects_size))
            alpha = theano.tensor.zeros(y.shape)
            for be in be_idx:
                a = []
                for i, b in enumerate(be):
                    a.append(batch_effects[:, i] == b)
                idx = reduce(np.logical_and, a).nonzero()
                if idx[0].shape[0] != 0:
                    alpha = theano.tensor.set_subtensor(alpha[idx, 0], skewness[be])
        else:
            alpha = 0  # symmetrical normal distribution

        y_like = pm.SkewNormal('y_like', mu=y_hat, sigma=sigma_y, alpha=alpha, observed=y)

    return model



#!/opt/conda/bin/python

# -----------------------------------------------------------------------------
# Run parallel normative modelling.
# All processing takes place in the processing directory (processing_dir)
# All inputs should be text files or binaries and space seperated
#
# It is possible to run these functions using...
#
# * k-fold cross-validation
# * estimating a training dataset then applying to a second test dataset
#
# First,the data is split for parallel processing.
# Second, the splits are submitted to the cluster.
# Third, the output is collected and combined.
#
# witten by (primarily) T Wolfers, (adaptated) SM Kia, H Huijsdens, L Parks, 
# S Rutherford, AF Marquand
# -----------------------------------------------------------------------------

from __future__ import print_function
from __future__ import division

import os
import sys
import glob
import shutil
import pickle
import fileinput
import time
import numpy as np
import pandas as pd
from subprocess import call, check_output


try:
    import pcntoolkit as ptk
    import pcntoolkit.dataio.fileio as fileio
    from pcntoolkit import configs
    from pcntoolkit.util.utils import yes_or_no 
    ptkpath = ptk.__path__[0] 
except ImportError:
    pass
    ptkpath = os.path.abspath(os.path.dirname(__file__))
    if ptkpath not in sys.path:
        sys.path.append(ptkpath)
    import dataio.fileio as fileio
    import configs
    from util.utils import yes_or_no 
    
    
PICKLE_PROTOCOL = configs.PICKLE_PROTOCOL


def execute_nm(processing_dir,
               python_path,
               job_name,
               covfile_path,
               respfile_path,
               batch_size,
               memory,
               duration,
               normative_path=None,
               func='estimate',
               interactive=False,
               **kwargs):

    ''' Execute parallel normative models
    This function is a mother function that executes all parallel normative
    modelling routines. Different specifications are possible using the sub-
    functions.

    Basic usage::

        execute_nm(processing_dir, python_path, job_name, covfile_path, respfile_path, batch_size, memory, duration)

    :param processing_dir: Full path to the processing dir
    :param python_path: Full path to the python distribution
    :param normative_path: Full path to the normative.py. If None (default) then it will automatically retrieves the path from the installed packeage.
    :param job_name: Name for the bash script that is the output of this function
    :param covfile_path: Full path to a .txt file that contains all covariats (subjects x covariates) for the responsefile
    :param respfile_path: Full path to a .txt that contains all features (subjects x features)
    :param batch_size: Number of features in each batch
    :param memory: Memory requirements written as string for example 4gb or 500mb
    :param duation: The approximate duration of the job, a string with HH:MM:SS for example 01:01:01
    :param cv_folds: Number of cross validations
    :param testcovfile_path: Full path to a .txt file that contains all covariates (subjects x covariates) for the test response file
    :param testrespfile_path: Full path to a .txt file that contains all test features
    :param log_path: Path for saving log files
    :param binary: If True uses binary format for response file otherwise it is text
    :param interactive: If False (default) the user should manually 
                        rerun the failed jobs or collect the results.
                        If 'auto' the job status are checked until all 
                        jobs are completed then the failed jobs are rerun
                        and the results are automaticallu collectted.
                        Using 'query' is similar to 'auto' unless it
                        asks for user verification thius is immune to 
                        endless loop in the case of bugs in the code.

    written by (primarily) T Wolfers, (adapted) SM Kia
    The documentation is adapated by S Rutherford.
    '''
    
    if normative_path is None:
        normative_path = ptkpath + '/normative.py'
        
    cv_folds = kwargs.get('cv_folds', None)
    testcovfile_path = kwargs.get('testcovfile_path', None)
    testrespfile_path= kwargs.get('testrespfile_path', None)
    outputsuffix = kwargs.get('outputsuffix', 'estimate')
    cluster_spec = kwargs.pop('cluster_spec', 'torque')
    log_path = kwargs.get('log_path', None)
    binary = kwargs.pop('binary', False)
    
    split_nm(processing_dir,
             respfile_path,
             batch_size,
             binary,
             **kwargs)

    batch_dir = glob.glob(processing_dir + 'batch_*')
    # print(batch_dir)
    number_of_batches = len(batch_dir)
    # print(number_of_batches)

    if binary:
        file_extentions = '.pkl'
    else:
        file_extentions = '.txt'
    
    kwargs.update({'batch_size':str(batch_size)})
    job_ids = []
    for n in range(1, number_of_batches+1):
        kwargs.update({'job_id':str(n)})
        if testrespfile_path is not None:
            if cv_folds is not None:
                raise(ValueError, """If the response file is specified
                                     cv_folds must be equal to None""")
            else:
                # specified train/test split
                batch_processing_dir = processing_dir + 'batch_' + str(n) + '/'
                batch_job_name = job_name + '_' + str(n) + '.sh'
                batch_respfile_path = (batch_processing_dir + 'resp_batch_' +
                                       str(n) + file_extentions)
                batch_testrespfile_path = (batch_processing_dir +
                                           'testresp_batch_' +
                                           str(n) + file_extentions)
                batch_job_path = batch_processing_dir + batch_job_name
                if cluster_spec == 'torque':
                    
                    # update the response file 
                    kwargs.update({'testrespfile_path': \
                                   batch_testrespfile_path})
                    bashwrap_nm(batch_processing_dir,
                                python_path,
                                normative_path,
                                batch_job_name,
                                covfile_path,
                                batch_respfile_path,
                                func=func,
                                **kwargs)
                    job_id = qsub_nm(job_path=batch_job_path,
                            log_path=log_path,
                            memory=memory,
                            duration=duration)
                    job_ids.append(job_id)
                elif cluster_spec == 'sbatch':
                    # update the response file 
                    kwargs.update({'testrespfile_path': \
                                   batch_testrespfile_path})
                    sbatchwrap_nm(batch_processing_dir,
                                python_path,
                                normative_path,
                                batch_job_name,
                                covfile_path,
                                batch_respfile_path,
                                func=func,
                                memory=memory,
                                duration=duration,
                                **kwargs)
                    sbatch_nm(job_path=batch_job_path,
                            log_path=log_path)
                elif cluster_spec == 'new':
                    # this part requires addition in different envioronment [
                    sbatchwrap_nm(processing_dir=batch_processing_dir, 
                                  func=func, **kwargs)
                    sbatch_nm(processing_dir=batch_processing_dir)
                    # ]
        if testrespfile_path is None:
            if testcovfile_path is not None:
                # forward model
                batch_processing_dir = processing_dir + 'batch_' + str(n) + '/'
                batch_job_name = job_name + '_' + str(n) + '.sh'
                batch_respfile_path = (batch_processing_dir + 'resp_batch_' +
                                       str(n) + file_extentions)
                batch_job_path = batch_processing_dir + batch_job_name
                if cluster_spec == 'torque':
                    bashwrap_nm(batch_processing_dir,
                                python_path,
                                normative_path,
                                batch_job_name,
                                covfile_path,
                                batch_respfile_path,
                                func=func,
                                **kwargs)
                    job_id = qsub_nm(job_path=batch_job_path,
                            log_path=log_path,
                            memory=memory,
                            duration=duration)
                    job_ids.append(job_id)
                elif cluster_spec == 'sbatch':
                    sbatchwrap_nm(batch_processing_dir,
                                python_path,
                                normative_path,
                                batch_job_name,
                                covfile_path,
                                batch_respfile_path,
                                func=func,
                                memory=memory,
                                duration=duration,
                                **kwargs)
                    sbatch_nm(job_path=batch_job_path,
                              log_path=log_path)
                elif cluster_spec == 'new':
                    # this part requires addition in different envioronment [
                    bashwrap_nm(processing_dir=batch_processing_dir, func=func,
                                **kwargs)
                    qsub_nm(processing_dir=batch_processing_dir)
                    # ]
            else:
                # cross-validation
                batch_processing_dir = (processing_dir + 'batch_' +
                                        str(n) + '/')
                batch_job_name = job_name + '_' + str(n) + '.sh'
                batch_respfile_path = (batch_processing_dir +
                                       'resp_batch_' + str(n) +
                                       file_extentions)
                batch_job_path = batch_processing_dir + batch_job_name
                if cluster_spec == 'torque':
                    bashwrap_nm(batch_processing_dir,
                                python_path,
                                normative_path,
                                batch_job_name,
                                covfile_path,
                                batch_respfile_path,
                                func=func,
                                **kwargs)
                    job_id = qsub_nm(job_path=batch_job_path,
                            log_path=log_path,
                            memory=memory,
                            duration=duration)
                    job_ids.append(job_id)
                elif cluster_spec == 'sbatch':
                    sbatchwrap_nm(batch_processing_dir,
                                python_path,
                                normative_path,
                                batch_job_name,
                                covfile_path,
                                batch_respfile_path,
                                func=func,
                                memory=memory,
                                duration=duration,
                                **kwargs)
                    sbatch_nm(job_path=batch_job_path,
                            log_path=log_path)
                elif cluster_spec == 'new':
                    # this part requires addition in different envioronment [
                    bashwrap_nm(processing_dir=batch_processing_dir, func=func,
                                **kwargs)
                    qsub_nm(processing_dir=batch_processing_dir)
                    # ]

    if interactive:
        
        check_jobs(job_ids, delay=60)
        
        success = False
        while (not success):
            success = collect_nm(processing_dir,
                           job_name,
                           func=func,
                           collect=False,
                           binary=binary,
                           batch_size=batch_size,
                           outputsuffix=outputsuffix)
            if success:
                break
            else:
                if interactive == 'query':
                    response = yes_or_no('Rerun the failed jobs?')
                    if response:
                        rerun_nm(processing_dir, log_path=log_path, memory=memory, 
                                 duration=duration, binary=binary, 
                                 interactive=interactive)
                    else:
                        success = True
                else:
                    print('Reruning the failed jobs ...')
                    rerun_nm(processing_dir, log_path=log_path, memory=memory, 
                                 duration=duration, binary=binary, 
                                 interactive=interactive)
                    
        if interactive == 'query':
            response = yes_or_no('Collect the results?')
            if response:
                success = collect_nm(processing_dir,
                               job_name,
                               func=func,
                               collect=True,
                               binary=binary,
                               batch_size=batch_size,
                               outputsuffix=outputsuffix)
        else:
            print('Collecting the results ...')
            success = collect_nm(processing_dir,
                               job_name,
                               func=func,
                               collect=True,
                               binary=binary,
                               batch_size=batch_size,
                               outputsuffix=outputsuffix)


"""routines that are environment independent"""

def split_nm(processing_dir,
             respfile_path,
             batch_size,
             binary,
             **kwargs):

    ''' This function prepares the input files for normative_parallel.
    
    Basic usage::

        split_nm(processing_dir, respfile_path, batch_size, binary, testrespfile_path)

    :param processing_dir: Full path to the processing dir
    :param respfile_path: Full path to the responsefile.txt (subjects x features)
    :param batch_size: Number of features in each batch
    :param testrespfile_path: Full path to the test responsefile.txt (subjects x features)
    :param binary: If True binary file

    :outputs: The creation of a folder struture for batch-wise processing.

    witten by (primarily) T Wolfers (adapted) SM Kia, (adapted) S Rutherford.
    ''' 
    
    testrespfile_path = kwargs.pop('testrespfile_path', None)

    dummy, respfile_extension = os.path.splitext(respfile_path)
    if (binary and respfile_extension != '.pkl'):
        raise(ValueError, """If binary is True the file format for the
              testrespfile file must be .pkl""")
    elif (binary==False and respfile_extension != '.txt'):
        raise(ValueError, """If binary is False the file format for the
              testrespfile file must be .txt""")

    # splits response into batches
    if testrespfile_path is None:
        if (binary==False):
            respfile = fileio.load_ascii(respfile_path)
        else:
            respfile = pd.read_pickle(respfile_path)

        respfile = pd.DataFrame(respfile)

        numsub = respfile.shape[1]
        batch_vec = np.arange(0,
                              numsub,
                              batch_size)
        batch_vec = np.append(batch_vec,
                              numsub)
        
        for n in range(0, (len(batch_vec) - 1)):
            resp_batch = respfile.iloc[:, (batch_vec[n]): batch_vec[n + 1]]
            os.chdir(processing_dir)
            resp = str('resp_batch_' + str(n+1))
            batch = str('batch_' + str(n+1))
            if not os.path.exists(processing_dir + batch):
                os.makedirs(processing_dir + batch)
                os.makedirs(processing_dir + batch + '/Models/')
            if (binary==False):
                fileio.save_pd(resp_batch,
                               processing_dir + batch + '/' +
                               resp + '.txt')
            else:
                resp_batch.to_pickle(processing_dir + batch + '/' +
                                     resp + '.pkl', protocol=PICKLE_PROTOCOL)

    # splits response and test responsefile into batches
    else:
        dummy, testrespfile_extension = os.path.splitext(testrespfile_path)
        if (binary and testrespfile_extension != '.pkl'):
            raise(ValueError, """If binary is True the file format for the
                  testrespfile file must be .pkl""")
        elif(binary==False and testrespfile_extension != '.txt'):
            raise(ValueError, """If binary is False the file format for the
                  testrespfile file must be .txt""")

        if (binary==False):
            respfile = fileio.load_ascii(respfile_path)
            testrespfile = fileio.load_ascii(testrespfile_path)
        else:
            respfile = pd.read_pickle(respfile_path)
            testrespfile = pd.read_pickle(testrespfile_path)

        respfile = pd.DataFrame(respfile)
        testrespfile = pd.DataFrame(testrespfile)

        numsub = respfile.shape[1]
        batch_vec = np.arange(0, numsub,
                              batch_size)
        batch_vec = np.append(batch_vec,
                              numsub)
        for n in range(0, (len(batch_vec) - 1)):
            resp_batch = respfile.iloc[:, (batch_vec[n]): batch_vec[n + 1]]
            testresp_batch = testrespfile.iloc[:, (batch_vec[n]): batch_vec[n +
                                             1]]
            os.chdir(processing_dir)
            resp = str('resp_batch_' + str(n+1))
            testresp = str('testresp_batch_' + str(n+1))
            batch = str('batch_' + str(n+1))
            if not os.path.exists(processing_dir + batch):
                os.makedirs(processing_dir + batch)
                os.makedirs(processing_dir + batch + '/Models/')
            if (binary==False):
                fileio.save_pd(resp_batch,
                               processing_dir + batch + '/' +
                               resp + '.txt')
                fileio.save_pd(testresp_batch,
                               processing_dir + batch + '/' + testresp +
                               '.txt')
            else:
                resp_batch.to_pickle(processing_dir + batch + '/' +
                                     resp + '.pkl', protocol=PICKLE_PROTOCOL)
                testresp_batch.to_pickle(processing_dir + batch + '/' +
                                         testresp + '.pkl', 
                                         protocol=PICKLE_PROTOCOL)


def collect_nm(processing_dir,
               job_name,
               func='estimate',
               collect=False,
               binary=False,
               batch_size=None,
               outputsuffix='_estimate'):
    
    '''Function to checks and collects all batches.

    Basic usage::

        collect_nm(processing_dir, job_name)


    :param processing_dir: Full path to the processing directory
    :param collect: If True data is checked for failed batches and collected; if False data is just checked
    :param binary: Results in pkl format

    :outputs: Text files containing all results accross all batches the combined output (written to disk).

    :returns 0: if batches fail
    :returns 1: if bathches complete successfully

    written by (primarily) T Wolfers, (adapted) SM Kia, (adapted) S Rutherford.
    '''

    if binary:
        file_extentions = '.pkl'
    else:
        file_extentions = '.txt'

    # detect number of subjects, batches, hyperparameters and CV
    batches = glob.glob(processing_dir + 'batch_*/')
    
    count = 0
    batch_fail = []
    
    if (func!='fit' and func!='extend' and func!='merge' and func!='tune'):
        file_example = []
        # TODO: Collect_nm only depends on yhat, thus does not work when no 
        # prediction is made (when test cov is not specified). 
        for batch in batches:
            if file_example == []:
                file_example = glob.glob(batch + 'yhat' + outputsuffix 
                                         + file_extentions)
            else:
                break
        if binary is False:
            file_example = fileio.load(file_example[0])
        else:
            file_example = pd.read_pickle(file_example[0])
        numsubjects = file_example.shape[0]
        batch_size = file_example.shape[1]
    
        # artificially creates files for batches that were not executed
        batch_dirs = glob.glob(processing_dir + 'batch_*/')
        batch_dirs = fileio.sort_nicely(batch_dirs)
        for batch in batch_dirs:
            filepath = glob.glob(batch + 'yhat' + outputsuffix + '*')
            if filepath == []:
                count = count+1
                batch1 = glob.glob(batch + '/' + job_name + '*.sh')
                print(batch1)
                batch_fail.append(batch1)
                if collect is True:
                    pRho = np.ones(batch_size)
                    pRho = pRho.transpose()
                    pRho = pd.Series(pRho)
                    fileio.save(pRho, batch + 'pRho' + outputsuffix + 
                                file_extentions)
                    
                    Rho = np.zeros(batch_size)
                    Rho = Rho.transpose()
                    Rho = pd.Series(Rho)
                    fileio.save(Rho, batch + 'Rho' + outputsuffix + 
                                file_extentions)
                    
                    rmse = np.zeros(batch_size)
                    rmse = rmse.transpose()
                    rmse = pd.Series(rmse)
                    fileio.save(rmse, batch + 'RMSE' + outputsuffix + 
                                file_extentions)
                    
                    smse = np.zeros(batch_size)
                    smse = smse.transpose()
                    smse = pd.Series(smse)
                    fileio.save(smse, batch + 'SMSE' + outputsuffix + 
                                file_extentions)
                    
                    expv = np.zeros(batch_size)
                    expv = expv.transpose()
                    expv = pd.Series(expv)
                    fileio.save(expv, batch + 'EXPV' + outputsuffix + 
                                file_extentions)
                    
                    msll = np.zeros(batch_size)
                    msll = msll.transpose()
                    msll = pd.Series(msll)
                    fileio.save(msll, batch + 'MSLL' + outputsuffix + 
                                file_extentions)
    
                    yhat = np.zeros([numsubjects, batch_size])
                    yhat = pd.DataFrame(yhat)
                    fileio.save(yhat, batch + 'yhat' + outputsuffix + 
                                file_extentions)
    
                    ys2 = np.zeros([numsubjects, batch_size])
                    ys2 = pd.DataFrame(ys2)
                    fileio.save(ys2, batch + 'ys2' + outputsuffix + 
                                file_extentions)
    
                    Z = np.zeros([numsubjects, batch_size])
                    Z = pd.DataFrame(Z)
                    fileio.save(Z, batch + 'Z' + outputsuffix + 
                                file_extentions)
                    
                    nll = np.zeros(batch_size)
                    nll = nll.transpose()
                    nll = pd.Series(nll)
                    fileio.save(nll, batch + 'NLL' + outputsuffix + 
                                file_extentions)
                    
                    bic = np.zeros(batch_size)
                    bic = bic.transpose()
                    bic = pd.Series(bic)
                    fileio.save(bic, batch + 'BIC' + outputsuffix + 
                                file_extentions)
    
                    if not os.path.isdir(batch + 'Models'):
                        os.mkdir('Models')
                        
                        
            else: # if more than 10% of yhat is nan then it is a failed batch
                yhat = fileio.load(filepath[0])
                if np.count_nonzero(~np.isnan(yhat))/(np.prod(yhat.shape))<0.9:
                    count = count+1
                    batch1 = glob.glob(batch + '/' + job_name + '*.sh')
                    print('More than 10% nans in '+ batch1[0])
                    batch_fail.append(batch1)
    
    else:
        batch_dirs = glob.glob(processing_dir + 'batch_*/')
        batch_dirs = fileio.sort_nicely(batch_dirs)
        for batch in batch_dirs:
            filepath = glob.glob(batch + 'Models/' + 'NM_' + '*' + outputsuffix 
                                 + '*')
            if len(filepath) < batch_size:
                count = count+1
                batch1 = glob.glob(batch + '/' + job_name + '*.sh')
                print(batch1)
                batch_fail.append(batch1)
    
    # combines all output files across batches
    if collect is True:
        pRho_filenames = glob.glob(processing_dir + 'batch_*/' + 'pRho' + 
                                   outputsuffix + '*')
        if pRho_filenames:
            pRho_filenames = fileio.sort_nicely(pRho_filenames)
            pRho_dfs = []
            for pRho_filename in pRho_filenames:
                pRho_dfs.append(pd.DataFrame(fileio.load(pRho_filename)))
            pRho_dfs = pd.concat(pRho_dfs, ignore_index=True, axis=0)
            fileio.save(pRho_dfs, processing_dir + 'pRho' + outputsuffix +
                        file_extentions)
            del pRho_dfs

        Rho_filenames = glob.glob(processing_dir + 'batch_*/' + 'Rho' + 
                                   outputsuffix + '*')
        if Rho_filenames:
            Rho_filenames = fileio.sort_nicely(Rho_filenames)
            Rho_dfs = []
            for Rho_filename in Rho_filenames:
                Rho_dfs.append(pd.DataFrame(fileio.load(Rho_filename)))
            Rho_dfs = pd.concat(Rho_dfs, ignore_index=True, axis=0)
            fileio.save(Rho_dfs, processing_dir + 'Rho' + outputsuffix +
                        file_extentions)
            del Rho_dfs

        Z_filenames = glob.glob(processing_dir + 'batch_*/' + 'Z' + 
                                   outputsuffix + '*')
        if Z_filenames:
            Z_filenames = fileio.sort_nicely(Z_filenames)
            Z_dfs = []
            for Z_filename in Z_filenames:
                Z_dfs.append(pd.DataFrame(fileio.load(Z_filename)))
            Z_dfs = pd.concat(Z_dfs, ignore_index=True, axis=1)
            fileio.save(Z_dfs, processing_dir + 'Z' + outputsuffix +
                        file_extentions)
            del Z_dfs
            
        yhat_filenames = glob.glob(processing_dir + 'batch_*/' + 'yhat' + 
                                   outputsuffix + '*')
        if yhat_filenames:
            yhat_filenames = fileio.sort_nicely(yhat_filenames)
            yhat_dfs = []
            for yhat_filename in yhat_filenames:
                yhat_dfs.append(pd.DataFrame(fileio.load(yhat_filename)))
            yhat_dfs = pd.concat(yhat_dfs, ignore_index=True, axis=1)
            fileio.save(yhat_dfs, processing_dir + 'yhat' + outputsuffix +
                        file_extentions)
            del yhat_dfs

        ys2_filenames = glob.glob(processing_dir + 'batch_*/' + 'ys2' + 
                                   outputsuffix + '*')
        if ys2_filenames:
            ys2_filenames = fileio.sort_nicely(ys2_filenames)
            ys2_dfs = []
            for ys2_filename in ys2_filenames:
                ys2_dfs.append(pd.DataFrame(fileio.load(ys2_filename)))
            ys2_dfs = pd.concat(ys2_dfs, ignore_index=True, axis=1)
            fileio.save(ys2_dfs, processing_dir + 'ys2' + outputsuffix +
                        file_extentions)
            del ys2_dfs

        rmse_filenames = glob.glob(processing_dir + 'batch_*/' + 'RMSE' + 
                                   outputsuffix + '*')
        if rmse_filenames:
            rmse_filenames = fileio.sort_nicely(rmse_filenames)
            rmse_dfs = []
            for rmse_filename in rmse_filenames:
                rmse_dfs.append(pd.DataFrame(fileio.load(rmse_filename)))
            rmse_dfs = pd.concat(rmse_dfs, ignore_index=True, axis=0)
            fileio.save(rmse_dfs, processing_dir + 'RMSE' + outputsuffix +
                        file_extentions)
            del rmse_dfs

        smse_filenames = glob.glob(processing_dir + 'batch_*/' + 'SMSE' + 
                                   outputsuffix + '*')
        if smse_filenames:
            smse_filenames = fileio.sort_nicely(smse_filenames)
            smse_dfs = []
            for smse_filename in smse_filenames:
                smse_dfs.append(pd.DataFrame(fileio.load(smse_filename)))
            smse_dfs = pd.concat(smse_dfs, ignore_index=True, axis=0)
            fileio.save(smse_dfs, processing_dir + 'SMSE' + outputsuffix +
                        file_extentions)
            del smse_dfs
            
        expv_filenames = glob.glob(processing_dir + 'batch_*/' + 'EXPV' + 
                                   outputsuffix + '*')
        if expv_filenames:
            expv_filenames = fileio.sort_nicely(expv_filenames)
            expv_dfs = []
            for expv_filename in expv_filenames:
                expv_dfs.append(pd.DataFrame(fileio.load(expv_filename)))
            expv_dfs = pd.concat(expv_dfs, ignore_index=True, axis=0)
            fileio.save(expv_dfs, processing_dir + 'EXPV' + outputsuffix +
                        file_extentions)
            del expv_dfs
            
        msll_filenames = glob.glob(processing_dir + 'batch_*/' + 'MSLL' + 
                                   outputsuffix + '*')
        if msll_filenames:
            msll_filenames = fileio.sort_nicely(msll_filenames)
            msll_dfs = []
            for msll_filename in msll_filenames:
                msll_dfs.append(pd.DataFrame(fileio.load(msll_filename)))
            msll_dfs = pd.concat(msll_dfs, ignore_index=True, axis=0)
            fileio.save(msll_dfs, processing_dir + 'MSLL' + outputsuffix +
                        file_extentions)
            del msll_dfs
            
        nll_filenames = glob.glob(processing_dir + 'batch_*/' + 'NLL' +
                                  outputsuffix + '*')
        if nll_filenames:
            nll_filenames = fileio.sort_nicely(nll_filenames)
            nll_dfs = []
            for nll_filename in nll_filenames:
                nll_dfs.append(pd.DataFrame(fileio.load(nll_filename)))
            nll_dfs = pd.concat(nll_dfs, ignore_index=True, axis=0)
            fileio.save(nll_dfs, processing_dir + 'NLL' + outputsuffix +
                        file_extentions)
            del nll_dfs

        bic_filenames = glob.glob(processing_dir + 'batch_*/' + 'BIC' +
                                  outputsuffix + '*')
        if bic_filenames:
            bic_filenames = fileio.sort_nicely(bic_filenames)
            bic_dfs = []
            for bic_filename in bic_filenames:
                bic_dfs.append(pd.DataFrame(fileio.load(bic_filename)))
            bic_dfs = pd.concat(bic_dfs, ignore_index=True, axis=0)
            fileio.save(bic_dfs, processing_dir + 'BIC' + outputsuffix +
                        file_extentions)
            del bic_dfs
        
        if (func!='predict' and func!='extend' and func!='merge' and func!='tune'):
            if not os.path.isdir(processing_dir + 'Models') and \
               os.path.exists(os.path.join(batches[0], 'Models')):
                os.mkdir(processing_dir + 'Models')
                
            meta_filenames = glob.glob(processing_dir + 'batch_*/Models/' + 
                                       'meta_data.md')
            mY = []
            sY = []
            X_scalers = []
            Y_scalers = []
            if meta_filenames:
                meta_filenames = fileio.sort_nicely(meta_filenames)
                with open(meta_filenames[0], 'rb') as file:
                    meta_data = pickle.load(file)
                
                for meta_filename in meta_filenames:
                    with open(meta_filename, 'rb') as file:
                        meta_data = pickle.load(file)
                    mY.append(meta_data['mean_resp'])
                    sY.append(meta_data['std_resp'])
                    if meta_data['inscaler'] in ['standardize', 'minmax', 
                                'robminmax']:
                        X_scalers.append(meta_data['scaler_cov'])
                    if meta_data['outscaler'] in ['standardize', 'minmax', 
                                'robminmax']:
                        Y_scalers.append(meta_data['scaler_resp'])
                meta_data['mean_resp'] = np.squeeze(np.column_stack(mY)) 
                meta_data['std_resp'] = np.squeeze(np.column_stack(sY))
                meta_data['scaler_cov'] = X_scalers 
                meta_data['scaler_resp'] = Y_scalers
                
                with open(os.path.join(processing_dir, 'Models', 
                                       'meta_data.md'), 'wb') as file:
                    pickle.dump(meta_data, file, protocol=PICKLE_PROTOCOL)
            
            batch_dirs = glob.glob(processing_dir + 'batch_*/')
            if batch_dirs:
                batch_dirs = fileio.sort_nicely(batch_dirs)
                for b, batch_dir in enumerate(batch_dirs):
                    src_files = glob.glob(batch_dir + 'Models/NM*' + 
                                          outputsuffix + '.pkl')
                    if src_files:
                        src_files = fileio.sort_nicely(src_files)
                        for f, full_file_name in enumerate(src_files):
                            if os.path.isfile(full_file_name):
                                file_name = full_file_name.split('/')[-1]
                                n = file_name.split('_')
                                n[-2] = str(b * batch_size + f)
                                n = '_'.join(n)
                                shutil.copy(full_file_name, processing_dir + 
                                            'Models/' + n)
                    elif func=='fit':
                        count = count+1
                        batch1 = glob.glob(batch_dir + '/' + job_name + '*.sh')
                        print('Failed batch: ' + batch1[0])
                        batch_fail.append(batch1)
                        
    # list batches that were not executed
    print('Number of batches that failed:' + str(count))
    batch_fail_df = pd.DataFrame(batch_fail)
    if file_extentions == '.txt':
        fileio.save_pd(batch_fail_df, processing_dir + 'failed_batches'+
                file_extentions)
    else:
        fileio.save(batch_fail_df, processing_dir +
            'failed_batches' +
            file_extentions)

    if not batch_fail:
        return True
    else:
        return False

def delete_nm(processing_dir,
              binary=False):
    '''This function deletes all processing for normative modelling and just keeps the combined output.

    Basic usage::

        collect_nm(processing_dir)

    :param processing_dir: Full path to the processing directory.
    :param binary: Results in pkl format.

    written by (primarily) T Wolfers, (adapted) SM Kia, (adapted) S Rutherford.
    '''
    
    if binary:
        file_extentions = '.pkl'
    else:
        file_extentions = '.txt'
    for file in glob.glob(processing_dir + 'batch_*/'):
        shutil.rmtree(file)
    if os.path.exists(processing_dir + 'failed_batches' + file_extentions):
        os.remove(processing_dir + 'failed_batches' + file_extentions)


# all routines below are envronment dependent and require adaptation in novel
# environments -> copy those routines and adapt them in accrodance with your
# environment

def bashwrap_nm(processing_dir,
                python_path,
                normative_path,
                job_name,
                covfile_path,
                respfile_path,
                func='estimate',
                **kwargs):

    ''' This function wraps normative modelling into a bash script to run it
    on a torque cluster system.

    Basic usage::

        bashwrap_nm(processing_dir, python_path, normative_path, job_name, covfile_path, respfile_path)

    :param processing_dir: Full path to the processing dir
    :param python_path: Full path to the python distribution
    :param normative_path: Full path to the normative.py
    :param job_name: Name for the bash script that is the output of this function
    :param covfile_path: Full path to a .txt file that contains all covariates (subjects x covariates) for the responsefile
    :param respfile_path: Full path to a .txt that contains all features (subjects x features)
    :param cv_folds: Number of cross validations
    :param testcovfile_path: Full path to a .txt file that contains all covariates (subjects x covariates) for the testresponse file
    :param testrespfile_path: Full path to a .txt file that contains all test features
    :param alg: which algorithm to use
    :param configparam: configuration parameters for this algorithm

    :outputs: A bash.sh file containing the commands for normative modelling saved to the processing directory (written to disk).

    written by (primarily) T Wolfers, (adapted) S Rutherford.
    '''
    
    # here we use pop not get to remove the arguments as they used 
    cv_folds = kwargs.pop('cv_folds',None)
    testcovfile_path = kwargs.pop('testcovfile_path', None)
    testrespfile_path = kwargs.pop('testrespfile_path', None)
    alg = kwargs.pop('alg', None)
    configparam = kwargs.pop('configparam', None)
    # change to processing dir
    os.chdir(processing_dir)
    output_changedir = ['cd ' + processing_dir + '\n']

    bash_lines = '#!/bin/bash\n'
    bash_cores = 'export OMP_NUM_THREADS=1\n'
    bash_environment = [bash_lines + bash_cores]

    # creates call of function for normative modelling
    if (testrespfile_path is not None) and (testcovfile_path is not None):
        job_call = [python_path + ' ' + normative_path + ' -c ' +
                    covfile_path + ' -t ' + testcovfile_path + ' -r ' +
                    testrespfile_path + ' -f ' + func]
    elif (testrespfile_path is None) and (testcovfile_path is not None):
        job_call = [python_path + ' ' + normative_path + ' -c ' +
                    covfile_path + ' -t ' + testcovfile_path + ' -f ' + func]
    elif cv_folds is not None:
        job_call = [python_path + ' ' + normative_path + ' -c ' +
                    covfile_path + ' -k ' + str(cv_folds) +  ' -f ' + func]
    elif func != 'estimate':
        job_call = [python_path + ' ' + normative_path + ' -c ' +
                    covfile_path +  ' -f ' + func]
    else:
        raise(ValueError, """For 'estimate' function either testcov or cvfold
              must be specified.""")
        
    # add algorithm-specific parameters
    if alg is not None:
        job_call = [job_call[0] + ' -a ' + alg]
        if configparam is not None:
            job_call = [job_call[0] + ' -x ' + str(configparam)]
    
    # add standardization flag if it is false
    # if not standardize:
    #     job_call = [job_call[0] + ' -s']
    
    # add responses file
    job_call = [job_call[0] + ' ' + respfile_path]
    
    # add in optional arguments. 
    for k in kwargs:
        job_call = [job_call[0] + ' ' + k + '=' + str(kwargs[k])]

    # writes bash file into processing dir
    with open(processing_dir+job_name, 'w') as bash_file:
        bash_file.writelines(bash_environment + output_changedir + \
                             job_call + ["\n"])

    # changes permissoins for bash.sh file
    os.chmod(processing_dir + job_name, 0o700)


def qsub_nm(job_path,
            log_path,
            memory,
            duration):
    
    '''This function submits a job.sh scipt to the torque custer using the qsub command.
    
    Basic usage::


        qsub_nm(job_path, log_path, memory, duration)

    :param job_path: Full path to the job.sh file.
    :param memory: Memory requirements written as string for example 4gb or 500mb.
    :param duation: The approximate duration of the job, a string with HH:MM:SS for example 01:01:01.

    :outputs: Submission of the job to the (torque) cluster.

    written by (primarily) T Wolfers, (adapted) SM Kia, (adapted) S Rutherford.
    '''
  
    # created qsub command
    if log_path is None:
        qsub_call = ['echo ' + job_path + ' | qsub -N ' + job_path + ' -l ' +
                     'procs=1' + ',mem=' + memory + ',walltime=' + duration]
    else:
        qsub_call = ['echo ' + job_path + ' | qsub -N ' + job_path +
                     ' -l ' + 'procs=1' + ',mem=' + memory + ',walltime=' + 
                     duration + ' -o ' + log_path + ' -e ' + log_path]

    # submits job to cluster
    #call(qsub_call, shell=True)
    job_id = check_output(qsub_call, shell=True).decode(sys.stdout.encoding).replace("\n", "")
    
    return job_id


def rerun_nm(processing_dir,
             log_path,
             memory,
             duration,
             binary=False, 
             interactive=False):
    '''This function reruns all failed batched in processing_dir after collect_nm has identified the failed batches.
    Basic usage::           

        rerun_nm(processing_dir, log_path, memory, duration)

    :param processing_dir: Full path to the processing directory
    :param memory: Memory requirements written as string for example 4gb or 500mb.
    :param duration: The approximate duration of the job, a string with HH:MM:SS for example 01:01:01.

    written by (primarily) T Wolfers, (adapted) SM Kia, (adapted) S Rutherford.
    '''
    
    job_ids = []
    

    if binary:
        file_extentions = '.pkl'
        failed_batches = fileio.load(processing_dir +
                                       'failed_batches' + file_extentions)
        shape = failed_batches.shape
        for n in range(0, shape[0]):
            jobpath = failed_batches[n, 0]
            print(jobpath)
            job_id = qsub_nm(job_path=jobpath,
                    log_path=log_path,
                    memory=memory,
                    duration=duration)
            job_ids.append(job_id)
    else:
        file_extentions = '.txt'
        failed_batches = fileio.load_pd(processing_dir +
                                       'failed_batches' + file_extentions)
        shape = failed_batches.shape
        for n in range(0, shape[0]):
            jobpath = failed_batches.iloc[n, 0]
            print(jobpath)
            job_id = qsub_nm(job_path=jobpath,
                    log_path=log_path,
                    memory=memory,
                    duration=duration)
            job_ids.append(job_id)
            
    if interactive: 
        check_jobs(job_ids, delay=60)
        

# COPY the rotines above here and aadapt those to your cluster
# bashwarp_nm; qsub_nm; rerun_nm

def sbatchwrap_nm(processing_dir,
                  python_path,
                  normative_path,
                  job_name,
                  covfile_path,
                  respfile_path,
                  memory,
                  duration,
                  func='estimate',
                  **kwargs):

    '''This function wraps normative modelling into a bash script to run it
    on a torque cluster system.

    Basic usage::

        sbatchwrap_nm(processing_dir, python_path, normative_path, job_name, covfile_path, respfile_path, memory, duration)

    :param processing_dir: Full path to the processing dir
    :param python_path: Full path to the python distribution
    :param normative_path: Full path to the normative.py
    :param job_name: Name for the bash script that is the output of this function
    :param covfile_path: Full path to a .txt file that contains all covariates (subjects x covariates) for the responsefile
    :param respfile_path: Full path to a .txt that contains all features (subjects x features)
    :param cv_folds: Number of cross validations
    :param testcovfile_path: Full path to a .txt file that contains all covariates (subjects x covariates) for the testresponse file
    :param testrespfile_path: Full path to a .txt file that contains all test features
    :param alg: which algorithm to use
    :param configparam: configuration parameters for this algorithm

    :outputs: A bash.sh file containing the commands for normative modelling saved to the processing directory (written to disk).

    written by (primarily) T Wolfers, (adapted) S Rutherford
    '''
    
    # here we use pop not get to remove the arguments as they used 
    cv_folds = kwargs.pop('cv_folds',None)
    testcovfile_path = kwargs.pop('testcovfile_path', None)
    testrespfile_path = kwargs.pop('testrespfile_path', None)
    alg = kwargs.pop('alg', None)
    configparam = kwargs.pop('configparam', None)
    
    # change to processing dir
    os.chdir(processing_dir)
    output_changedir = ['cd ' + processing_dir + '\n']

    sbatch_init='#!/bin/bash\n'
    sbatch_jobname='#SBATCH --job-name=' + processing_dir + '\n'
    sbatch_account='#SBATCH --account=p33_norment\n'
    sbatch_nodes='#SBATCH --nodes=1\n'
    sbatch_tasks='#SBATCH --ntasks=1\n'
    sbatch_time='#SBATCH --time=' + str(duration) + '\n'
    sbatch_memory='#SBATCH --mem-per-cpu=' + str(memory) + '\n'
    sbatch_module='module purge\n'
    sbatch_anaconda='module load anaconda3\n'
    sbatch_exit='set -o errexit\n'

    #echo -n "This script is running on "
    #hostname
    
    bash_environment = [sbatch_init + 
                        sbatch_jobname +
                        sbatch_account +
                        sbatch_nodes +
                        sbatch_tasks +
                        sbatch_time +
                        sbatch_module +
                        sbatch_anaconda]

    # creates call of function for normative modelling
    if (testrespfile_path is not None) and (testcovfile_path is not None):
        job_call = [python_path + ' ' + normative_path + ' -c ' +
                    covfile_path + ' -t ' + testcovfile_path + ' -r ' +
                    testrespfile_path + ' -f ' + func]
    elif (testrespfile_path is None) and (testcovfile_path is not None):
        job_call = [python_path + ' ' + normative_path + ' -c ' +
                    covfile_path + ' -t ' + testcovfile_path + ' -f ' + func]
    elif cv_folds is not None:
        job_call = [python_path + ' ' + normative_path + ' -c ' +
                    covfile_path + ' -k ' + str(cv_folds) +  ' -f ' + func]
    elif func != 'estimate':
        job_call = [python_path + ' ' + normative_path + ' -c ' +
                    covfile_path +  ' -f ' + func]
    else:
        raise(ValueError, """For 'estimate' function either testcov or cvfold
              must be specified.""")
        
    # add algorithm-specific parameters
    if alg is not None:
        job_call = [job_call[0] + ' -a ' + alg]
        if configparam is not None:
            job_call = [job_call[0] + ' -x ' + str(configparam)]
    
    # add standardization flag if it is false
    # if not standardize:
    #     job_call = [job_call[0] + ' -s']
    
    # add responses file
    job_call = [job_call[0] + ' ' + respfile_path]
    
    # add in optional arguments. 
    for k in kwargs:
        job_call = [job_call[0] + ' ' + k + '=' + kwargs[k]]

    # writes bash file into processing dir
    with open(processing_dir+job_name, 'w') as bash_file:
        bash_file.writelines(bash_environment + output_changedir + \
                             job_call + ["\n"] + [sbatch_exit])

    # changes permissoins for bash.sh file
    os.chmod(processing_dir + job_name, 0o700)

def sbatch_nm(job_path,
              log_path):
    
    '''This function submits a job.sh scipt to the torque custer using the qsub
    command.

    Basic usage::

        sbatch_nm(job_path, log_path)

    :param job_path: Full path to the job.sh file
    :param log_path: The logs are currently stored in the working dir

    :outputs: Submission of the job to the (torque) cluster.

    written by (primarily) T Wolfers, (adapted) S Rutherford.
    '''

    # created qsub command
    sbatch_call = ['sbatch ' + job_path]

    # submits job to cluster
    call(sbatch_call, shell=True)
    
def sbatchrerun_nm(processing_dir,
                 memory,
                 duration,
                 new_memory=False,
                 new_duration=False,
                 binary=False,
                 **kwargs):
        
    '''This function reruns all failed batched in processing_dir after collect_nm has identified he failed batches.
    
    Basic usage::

        rerun_nm(processing_dir, memory, duration)

    :param processing_dir: Full path to the processing directory.
    :param memory: Memory requirements written as string, for example 4gb or 500mb.
    :param duration: The approximate duration of the job, a string with HH:MM:SS for example 01:01:01.
    :param new_memory: If you want to change the memory you have to indicate it here.
    :param new_duration: If you want to change the duration you have to indicate it here.

    :outputs: Re-runs failed batches. 
    
     written by (primarily) T Wolfers, (adapted) S Rutherford.
    '''
    log_path = kwargs.pop('log_path', None)
    
    if binary:
        file_extentions = '.pkl'
        failed_batches = fileio.load(processing_dir + 'failed_batches' +  file_extentions)
        shape = failed_batches.shape
        for n in range(0, shape[0]):
            jobpath = failed_batches[n, 0]
            print(jobpath)
            if new_duration != False:
                with fileinput.FileInput(jobpath, inplace=True) as file:
                    for line in file:
                        print(line.replace(duration, new_duration), end='')
                if new_memory != False:
                    with fileinput.FileInput(jobpath, inplace=True) as file:
                        for line in file:
                            print(line.replace(memory, new_memory), end='')
                sbatch_nm(jobpath, log_path)

    else:
        file_extentions = '.txt'
        failed_batches = fileio.load_pd(processing_dir + 'failed_batches' + file_extentions)
        shape = failed_batches.shape
        for n in range(0, shape[0]):
            jobpath = failed_batches.iloc[n, 0]
            print(jobpath)
            if new_duration != False:
                with fileinput.FileInput(jobpath, inplace=True) as file:
                    for line in file:
                        print(line.replace(duration, new_duration), end='')
                if new_memory != False:
                    with fileinput.FileInput(jobpath, inplace=True) as file:
                        for line in file:
                            print(line.replace(memory, new_memory), end='')
                sbatch_nm(jobpath,
                          log_path)


def retrieve_jobs():
    """
    A utility function to retrieve task status from the outputs of qstat.
    
    :return: a dictionary of jobs.

    """
    
    output = check_output('qstat', shell=True).decode(sys.stdout.encoding)
    output = output.split('\n')
    jobs = dict()
    for line in output[2:-1]:
        (Job_ID, Job_Name, User, Wall_Time, Status, Queue) = line.split()
        jobs[Job_ID] = dict()
        jobs[Job_ID]['name'] = Job_Name
        jobs[Job_ID]['walltime'] = Wall_Time
        jobs[Job_ID]['status'] = Status
        
    return jobs
        

def check_job_status(jobs):
    """
    A utility function to count the tasks with different status.
    
    :param jobs: List of job ids.
    :return: returns the number of taks athat are queued, running, completed etc

    """
    running_jobs = retrieve_jobs()
    
    r = 0
    c = 0
    q = 0
    u = 0
    for job in jobs:
        try:
            if running_jobs[job]['status'] == 'C':
                c += 1
            elif running_jobs[job]['status'] == 'Q':
                q += 1
            elif running_jobs[job]['status'] == 'R':
                r += 1
            else:
                u += 1
        except: # probably meanwhile the job is finished.
            c += 1 
            continue
                 
    print('Total Jobs:%d, Queued:%d, Running:%d, Completed:%d, Unknown:%d' 
          %(len(jobs), q, r, c, u))
    return q,r,c,u
    

def check_jobs(jobs, delay=60):
    """
    A utility function for chacking the status of submitted jobs.
    
    :param jobs: list of job ids.
    :param delay: the delay (in sec) between two consequative checks, defaults to 60.

    """
    
    n = len(jobs)
    
    while(True):
        q,r,c,u = check_job_status(jobs)
        if c == n:
            print('All jobs are completed!')
            break
        time.sleep(delay)
        




from __future__ import print_function
from __future__ import division

import os
import sys
import numpy as np

try:  # run as a package if installed
    from pcntoolkit.normative_model.norm_base import NormBase
    from pcntoolkit.model.rfa import GPRRFA 
except ImportError:
    pass

    path = os.path.abspath(os.path.dirname(__file__))
    if path not in sys.path:
        sys.path.append(path)
    del path

    from model.rfa import GPRRFA
    from norm_base import NormBase

class NormRFA(NormBase):
    """ Classical GPR-based normative modelling approach
    """

    def __init__(self, X, y=None, theta=None, n_feat=None):
                        
        if (X is not None):
            if n_feat is None:
                print("initialising RFA")
            else:
                print("initialising RFA with", n_feat, "random features")
            self.gprrfa = GPRRFA(theta, X, n_feat=n_feat)
            self._n_params = self.gprrfa.get_n_params(X)
        else:
            raise(ValueError, 'please specify covariates')
            return
        
        if theta is None:
            self.theta0 = np.zeros(self._n_params)
        else:
            if len(theta) == self._n_params:
                self.theta0 = theta
            else:
                raise(ValueError, 'hyperparameter vector has incorrect size')
       
        self.theta = self.theta0
            
    @property
    def n_params(self):
           
        return self._n_params
    
    @property
    def neg_log_lik(self):
        return self.gprrfa.nlZ

    def estimate(self, X, y, theta=None):
        if theta is None:
            theta = self.theta0
        self.gprrfa = GPRRFA(theta, X, y)
        self.theta = self.gprrfa.estimate(theta, X, y)
        
        return self

    def predict(self, Xs, X, y, theta=None):
        if theta is None:
            theta = self.theta
        yhat, s2 = self.gprrfa.predict(theta, X, y, Xs)
        
        return yhat, s2
    


from __future__ import print_function
import os
import sys
import numpy as np
from scipy import stats
import scipy.special as spp
import pickle
import matplotlib.pyplot as plt
import pandas as pd
import pymc3 as pm
from pcntoolkit.model.SHASH import *
from pcntoolkit.model.hbr import bspline_transform

"""
@author: augub
"""

def MCMC_estimate(f, trace):
    """Get an MCMC estimate of f given a trace"""
    out = np.zeros_like(f(trace.point(0)))
    n=0
    for p in trace.points():
        out += f(p)
        n+=1
    return out/n


def get_MCMC_zscores(X, Y, Z, model):
    """Get an MCMC estimate of the z-scores of Y"""
    def f(sample):
         return get_single_zscores(X, Y, Z, model,sample)
    return MCMC_estimate(f, model.hbr.trace)


def get_single_zscores(X, Y, Z, model, sample):
    """Get the z-scores of y, given clinical covariates and a model"""
    likelihood = model.configs['likelihood']
    params = forward(X,Z,model,sample)
    return z_score(Y, params, likelihood = likelihood)
    

def z_score(Y, params, likelihood = "Normal"):
    """Get the z-scores of Y, given likelihood parameters"""
    if likelihood.startswith('SHASH'):
        mu = params['mu']
        sigma = params['sigma']
        epsilon = params['epsilon']
        delta = params['delta']
        if likelihood == "SHASHo":
            SHASH = (Y-mu)/sigma
            Z = np.sinh(np.arcsinh(SHASH)*delta - epsilon)
        elif likelihood == "SHASHo2":
            sigma_d = sigma/delta
            SHASH = (Y-mu)/sigma_d
            Z = np.sinh(np.arcsinh(SHASH)*delta - epsilon)
        elif likelihood == "SHASHb":
            true_mu = m(epsilon, delta, 1)
            true_sigma = np.sqrt((m(epsilon, delta, 2) - true_mu ** 2))
            SHASH_c = ((Y-mu)/sigma)
            SHASH = SHASH_c * true_sigma + true_mu
            Z = np.sinh(np.arcsinh(SHASH) * delta - epsilon)
    elif likelihood == 'Normal':
        Z = (Y-params['mu'])/params['sigma']
    else:
        exit("Unsupported likelihood")
    return Z


def get_MCMC_quantiles(synthetic_X, z_scores, model, be):
    """Get an MCMC estimate of the quantiles"""
    """This does not use the get_single_quantiles function, for memory efficiency"""
    resolution = synthetic_X.shape[0]
    synthetic_X_transformed = model.hbr.transform_X(synthetic_X)
    be = np.reshape(np.array(be),(1,-1))
    synthetic_Z = np.repeat(be, resolution, axis = 0)
    z_scores = np.reshape(np.array(z_scores),(1,-1))
    zs = np.repeat(z_scores, resolution, axis=0)
    def f(sample):
        params = forward(synthetic_X_transformed,synthetic_Z, model,sample)
        q = quantile(zs, params, likelihood = model.configs['likelihood'])
        return q
    out = MCMC_estimate(f, model.hbr.trace)
    return out


def get_single_quantiles(synthetic_X, z_scores, model, be, sample):
    """Get the quantiles within a given range of covariates, given a model"""
    resolution = synthetic_X.shape[0]
    synthetic_X_transformed = model.hbr.transform_X(synthetic_X)
    be = np.reshape(np.array(be),(1,-1))
    synthetic_Z = np.repeat(be, resolution, axis = 0)
    z_scores = np.reshape(np.array(z_scores),(1,-1))
    zs = np.repeat(z_scores, resolution, axis=0)
    params = forward(synthetic_X_transformed,synthetic_Z, model,sample)
    q = quantile(zs, params, likelihood = model.configs['likelihood'])
    return q


def quantile(zs, params, likelihood = "Normal"):
    """Get the zs'th quantiles given likelihood parameters"""
    if likelihood.startswith('SHASH'):
        mu = params['mu']
        sigma = params['sigma']
        epsilon = params['epsilon']
        delta = params['delta']
        if likelihood == "SHASHo":
            quantiles = S_inv(zs,epsilon,delta)*sigma + mu
        elif likelihood == "SHASHo2":
            sigma_d = sigma/delta
            quantiles = S_inv(zs,epsilon,delta)*sigma_d + mu
        elif likelihood == "SHASHb":
            true_mu = m(epsilon, delta, 1)
            true_sigma = np.sqrt((m(epsilon, delta, 2) - true_mu ** 2))
            SHASH_c = ((S_inv(zs,epsilon,delta)-true_mu)/true_sigma)
            quantiles = SHASH_c *sigma + mu
    elif likelihood == 'Normal':
        quantiles = zs*params['sigma'] + params['mu']
    else:
        exit("Unsupported likelihood")
    return quantiles


def single_parameter_forward(X, Z, model, sample, p_name):
    """Get a likelihood paramameter given covariates, batch-effects and model parameters"""
    outs = np.zeros(X.shape[0])[:,None]
    all_bes = np.unique(Z,axis=0)
    for be in all_bes:
        bet = tuple(be)
        idx = (Z==be).all(1)
        if model.configs[f"linear_{p_name}"]:
            if model.configs[f'random_slope_{p_name}']:
                slope_be = sample[f"slope_{p_name}"][bet]
            else:
                slope_be = sample[f"slope_{p_name}"]
            if model.configs[f'random_intercept_{p_name}']:
                intercept_be = sample[f"intercept_{p_name}"][bet]
            else:
                intercept_be = sample[f"intercept_{p_name}"]

            out = (X[np.squeeze(idx),:]@slope_be)[:,None] + intercept_be
            outs[np.squeeze(idx),:] = out
        else:
            if model.configs[f'random_{p_name}']:
                outs[np.squeeze(idx),:] = sample[p_name][bet]
            else:
                outs[np.squeeze(idx),:] = sample[p_name]

    return outs


def forward(X, Z, model, sample):
    """Get all likelihood paramameters given covariates and batch-effects and model parameters"""
    # TODO think if this is the correct spot for this
    mapfuncs={'sigma': lambda x: np.log(1+np.exp(x)), 'delta':lambda x :np.log(1+np.exp(x)) + 0.3}

    likelihood = model.configs['likelihood']

    if likelihood == 'Normal':
        parameter_list = ['mu','sigma']
    elif likelihood in ['SHASHb','SHASHo','SHASHo2']:
        parameter_list = ['mu','sigma','epsilon','delta']
    else:
        exit("Unsupported likelihood")

    for i in parameter_list:
        if not (i in mapfuncs.keys()):
            mapfuncs[i] = lambda x: x

    output_dict = {p_name:np.zeros(X.shape) for p_name in parameter_list}

    for p_name in parameter_list:
        output_dict[p_name] = mapfuncs[p_name](single_parameter_forward(X,Z,model,sample,p_name))

    return output_dict


def Rhats(model, thin = 1, resolution = 100, varnames = None):
    """Get Rhat as function of sampling iteration"""
    trace = model.hbr.trace

    if varnames == None:
        varnames = trace.varnames
    chain_length = trace.get_values(varnames[0],chains=trace.chains[0], thin=thin).shape[0]
    
    interval_skip=chain_length//resolution

    rhat_dict = {}

    for varname in varnames:
        testvar = np.stack(trace.get_values(varname,combine=False))
        vardim = testvar.reshape((testvar.shape[0], testvar.shape[1], -1)).shape[2]
        rhats_var = np.zeros((resolution, vardim))

        var = np.stack(trace.get_values(varname,combine=False))
        var = var.reshape((var.shape[0], var.shape[1], -1))    
        for v in range(var.shape[2]):
            for j in range(resolution):
                rhats_var[j,v] = pm.rhat(var[:,:j*interval_skip,v])
        rhat_dict[varname] = rhats_var
    return rhat_dict


def S_inv(x, e, d):
    return np.sinh((np.arcsinh(x) + e) / d)
    
def K(p, x):
    return np.array(spp.kv(p, x))

def P(q):
    """
    The P function as given in Jones et al.
    :param q:
    :return:
    """
    frac = np.exp(1 / 4) / np.sqrt(8 * np.pi)
    K1 = K((q + 1) / 2, 1 / 4)
    K2 = K((q - 1) / 2, 1 / 4)
    a = (K1 + K2) * frac
    return a

def m(epsilon, delta, r):
    """
    The r'th uncentered moment. Given by Jones et al.
    """
    frac1 = 1 / np.power(2, r)
    acc = 0
    for i in range(r + 1):
        combs = spp.comb(r, i)
        flip = np.power(-1, i)
        ex = np.exp((r - 2 * i) * epsilon / delta)
        p = P((r - 2 * i) / delta)
        acc += combs * flip * ex * p
    return frac1 * acc



